# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic to validate submission metadata based on a LinkML model."""

from typing import Any

from linkml_validator.models import ValidationMessage
from linkml_validator.validator import Validator

from metldata.model_utils.essentials import MetadataModel


class InvalidMetadataError(RuntimeError):
    """Raised when the metadata is invalid."""


class MetadataValidationError(InvalidMetadataError):
    """Raised when the validation of metadata against the provided model fails."""

    def __init__(self, *, issues: list[ValidationMessage]):
        """Initializes with list of issues."""

        self.issues = issues

        message = "Validation failed due to following issues: " + ", ".join(
            (
                f"in field '{issue.field}' with value '{issue.value}': {issue.message}"
                + f" ({issue.severity})"
            )
            for issue in self.issues
        )
        super().__init__(message)


class MetadataValidator:
    """Validating metadata against a LinkML model."""

    # error shortcuts:
    ValidationError = MetadataValidationError

    def __init__(self, *, model: MetadataModel):
        """Initialize the validator with a metadata model."""

        self._model = model

    def validate(self, metadata: dict[str, Any]) -> None:
        """Validate metadata against the provided model.

        Raises:
            ValidationError: When validation failed.
        """

        with self._model.temporary_yaml_path() as model_path:
            validator = Validator(schema=str(model_path))
            validation_report = validator.validate(metadata, target_class="Submission")

            if not validation_report.valid:
                issues = [
                    message
                    for result in validation_report.validation_results
                    if not result.valid and result.validation_messages is not None
                    for message in result.validation_messages
                ]
                raise MetadataValidationError(issues=issues)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for loading artifact resources."""

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.load_resources import (
    extract_all_resources_from_artifact,
    process_new_or_changed_resources,
    process_removed_resources,
)
from metldata.artifacts_rest.models import ArtifactInfo, ArtifactResource
from metldata.load.event_publisher import EventPublisherPort
from metldata.load.models import ArtifactResourceDict


class ArtifactResourcesInvalid(RuntimeError):
    """Raised when artifact resources are invalid."""


def check_artifact_resources(
    artifact_resources: ArtifactResourceDict, artifact_infos: dict[str, ArtifactInfo]
) -> None:
    """Check the provided artifact resources against the given artifact info.

    Raises:
        ArtifactResourcesInvalid: If the artifact resources are invalid.
    """

    for artifact_name in artifact_resources:
        if artifact_name not in artifact_infos:
            raise ArtifactResourcesInvalid(f"Artifact '{artifact_name}' is unknown.")


async def load_artifacts_using_dao(
    artifact_resources: ArtifactResourceDict,
    artifact_info_dict: dict[str, ArtifactInfo],
    event_publisher: EventPublisherPort,
    dao_collection: ArtifactDaoCollection,
) -> None:
    """Load artifact resources from multiple submissions using the given dao collection."""

    (
        removed_resource_tags,
        new_resources,
        changed_resources,
    ) = await _get_changed_resources(
        artifact_resources=artifact_resources,
        artifact_info_dict=artifact_info_dict,
        dao_collection=dao_collection,
    )

    await process_removed_resources(
        event_publisher=event_publisher,
        resource_tags=removed_resource_tags,
        dao_collection=dao_collection,
    )

    await process_new_or_changed_resources(
        artifact_info_dict=artifact_info_dict,
        event_publisher=event_publisher,
        resources=new_resources,
        dao_collection=dao_collection,
    )

    await process_new_or_changed_resources(
        artifact_info_dict=artifact_info_dict,
        event_publisher=event_publisher,
        resources=changed_resources,
        dao_collection=dao_collection,
    )


async def _get_changed_resources(  # pylint: disable=too-many-locals
    artifact_resources: ArtifactResourceDict,
    artifact_info_dict: dict[str, ArtifactInfo],
    dao_collection: ArtifactDaoCollection,
) -> tuple[
    set[tuple[str, str, str]],
    dict[tuple[str, str, str], ArtifactResource],
    dict[tuple[str, str, str], ArtifactResource],
]:
    """Extract changed resources using DAOs and currently submitted artifacts

    Returns the following collections in order:
     - A set of resource tags for artifact resources that have been removed
     - A dict of new artifact resources indexed by the corresponding resource_tag
     - A dict of changed artifact resources indexed by the corresponding resource_tag
    """

    # only need to collect resource tag to check if something needs to be deleted
    unchanged_resource_tags = []

    # collect new/changed resources indexed by their resource_tag for insert/upsert
    new_resources = {}
    changed_resources = {}

    existing_resource_tags = await dao_collection.get_all_resource_tags()

    for artifact_name, artifact_contents in artifact_resources.items():
        for artifact_content in artifact_contents:
            resources = extract_all_resources_from_artifact(
                artifact_content=artifact_content,
                artifact_info=artifact_info_dict[artifact_name],
            )

            # check for each resource if it does already exist and is changed
            for resource in resources:
                resource_tag = (artifact_name, resource.class_name, resource.id_)

                if resource_tag in existing_resource_tags:
                    dao = await dao_collection.get_dao(
                        artifact_name=artifact_name, class_name=resource.class_name
                    )
                    old_resource = await dao.get_by_id(resource.id_)

                    if old_resource == resource:
                        unchanged_resource_tags.append(resource_tag)
                    else:
                        changed_resources[resource_tag] = resource
                else:
                    new_resources[resource_tag] = resource

    removed_resource_tags = existing_resource_tags.difference(
        unchanged_resource_tags, new_resources, changed_resources
    )

    return removed_resource_tags, new_resources, changed_resources



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utilities for handling metadata."""

from copy import deepcopy
from typing import Union, cast

from metldata.custom_types import Json
from metldata.model_utils.anchors import AnchorPoint, lookup_anchor_point


class SelfIdLookUpError(RuntimeError):
    """Raised when the self id cannot be looked up."""


class ForeignIdLookUpError(RuntimeError):
    """Raised when a foreign id cannot be looked up."""


class MetadataAnchorMismatchError(RuntimeError):
    """Raised when the provided metadata does not match the expected anchor points."""


class MetadataResourceNotFoundError(RuntimeError):
    """Raised when a resource could not be found in the metadata."""


class IdenitifierConflictError(RuntimeError):
    """Raised when multiple resources of the same class with the same identifier exist."""


class SlotNotFoundError(RuntimeError):
    """Raised when a slot cannot be found in a Metadata Resource."""


def lookup_self_id(*, resource: Json, identifier_slot: str):
    """Lookup the ID of the specified resource."""

    self_id = resource.get(identifier_slot)

    if self_id is None:
        raise SelfIdLookUpError(
            "Cannot lookup the identifier because the corresponding slot"
            + f" '{identifier_slot}' is not present in the resource '{resource}'."
        )

    if not isinstance(self_id, str):
        raise SelfIdLookUpError(
            f"Cannot lookup the identifier because the slot '{identifier_slot}' of"
            + f" resource '{resource}' contains a non-string value."
        )

    return self_id


def lookup_foreign_ids(*, resource: Json, slot: str) -> set[str]:
    """Lookup foreign IDs referenced by the specified resource using the specified
    slot."""

    foreign_ids = resource.get(slot)

    if foreign_ids is None:
        raise ForeignIdLookUpError(
            f"Cannot lookup foreign IDs because the slot '{slot}' is not present"
            + f" in the resource '{resource}'."
        )

    # convert single value to list:
    if not isinstance(foreign_ids, list):
        foreign_ids = [foreign_ids]

    # check that all values are strings:
    if not all(isinstance(value, str) for value in foreign_ids):
        raise ForeignIdLookUpError(
            f"Cannot lookup foreign IDs because the slot '{slot}' of resource"
            + f" '{resource}' contains non-string values."
        )

    foreign_ids = set(foreign_ids)
    foreign_ids = cast(set[str], foreign_ids)

    return foreign_ids


def lookup_resource_by_identifier(
    *,
    class_name: str,
    identifier: str,
    global_metadata: Json,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Lookup a resource of the given class in the provided global metadata by its
    identifier.

    Raises:
        MetadataAnchorMismatchError:
            if the provided metadata does not match the expected anchor points.
        MetadataResourceNotFoundError:
            if the resource with the given identifier could not be found.
    """

    anchor_point = lookup_anchor_point(
        class_name=class_name, anchor_points_by_target=anchor_points_by_target
    )

    if anchor_point.root_slot not in global_metadata:
        raise MetadataAnchorMismatchError(
            f"Could not find root slot of the anchor point '{anchor_point.root_slot}'"
            + " in the global metadata."
        )

    resources = global_metadata[anchor_point.root_slot]

    resources_dict = convert_resource_list_to_dict(
        resources=resources, identifier_slot=anchor_point.identifier_slot
    )

    if identifier not in resources_dict:
        raise MetadataResourceNotFoundError(
            f"Could not find resource with identifier '{identifier}' of class"
            + f" '{class_name}' in the global metadata."
        )

    target_resource = resources_dict[identifier]

    return target_resource


def check_identifier_uniqueness(*, resources: list[Json], identifier_slot: str) -> None:
    """Check if the identifiers of the specified resources are unique.

    Raises:
        IdenitifierConflictError:
            if the identifiers are not unique.
    """

    identifiers = [resource[identifier_slot] for resource in resources]

    if len(identifiers) != len(set(identifiers)):
        raise IdenitifierConflictError(
            "The identifiers of the resources are not unique."
        )


def convert_list_to_inlined_dict(
    *, resources: list[Json], identifier_slot: str
) -> dict[str, Json]:
    """Convert a list of resources of same type into a dictionary representation, i.e.
    to "inlined_as_list=false" format.

    Raises:
        IdenitifierConflictError:
            if the identifiers are not unique.
    """

    check_identifier_uniqueness(resources=resources, identifier_slot=identifier_slot)

    return {
        lookup_self_id(resource=resource, identifier_slot=identifier_slot): {
            slot: deepcopy(resource[slot])
            for slot in resource
            if slot != identifier_slot
        }
        for resource in resources
    }


def convert_inlined_dict_to_list(
    *, resources: dict[str, Json], identifier_slot: str
) -> list[Json]:
    """Convert a dictionary representation of resources into a list representation, i.e.
    to "inlined_as_list=true" format."""

    resource_copy = deepcopy(resources)

    return [
        {identifier_slot: identifier, **resource_copy}
        for identifier, resource in resource_copy.items()
    ]


def convert_resource_list_to_dict(
    *, resources: list[Json], identifier_slot: str
) -> dict[str, Json]:
    """Convert a list of resources of same type into a dictionary representation.
    Unlike the `inlined_as_list=false` structure from LinkML, the dict will contain
    resources with identifier slots.

    Raises:

    """

    return {
        lookup_self_id(resource=resource, identifier_slot=identifier_slot): deepcopy(
            resource
        )
        for resource in resources
    }


def get_resources_of_class(
    *,
    class_name: str,
    global_metadata: Json,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> list[Json]:
    """Get all instances of the given class from the provided global metadata.

    Raises:
        MetadataAnchorMismatchError:
            if the provided metadata does not match the expected anchor points.
    """

    anchor_point = lookup_anchor_point(
        class_name=class_name, anchor_points_by_target=anchor_points_by_target
    )

    if anchor_point.root_slot not in global_metadata:
        raise MetadataAnchorMismatchError(
            f"Could not find root slot of the anchor point '{anchor_point.root_slot}'"
            + " in the global metadata."
        )

    return deepcopy(global_metadata[anchor_point.root_slot])


def get_resource_dict_of_class(
    *,
    class_name: str,
    global_metadata: Json,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> dict[str, Json]:
    """Get all instances as dict of the given class from the provided global metadata.
    Unlike the `inlined_as_list=false` structure from LinkML, the dict will contain
    resources with identifier slots.

    Raises:
        MetadataAnchorMismatchError:
            if the provided metadata does not match the expected anchor points.
    """

    resources = get_resources_of_class(
        class_name=class_name,
        global_metadata=global_metadata,
        anchor_points_by_target=anchor_points_by_target,
    )

    anchor_point = lookup_anchor_point(
        class_name=class_name, anchor_points_by_target=anchor_points_by_target
    )

    return convert_resource_list_to_dict(
        resources=resources, identifier_slot=anchor_point.identifier_slot
    )


def upsert_resources_in_metadata(
    *,
    resources: list[Json],
    class_name: str,
    global_metadata: Json,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Update the provided global metadata with the provided resources of the given
    class. If the anchor point for the given class does not yet exist, it is created.
    Returns a copy of the updated metadata."""

    anchor_point = lookup_anchor_point(
        class_name=class_name, anchor_points_by_target=anchor_points_by_target
    )

    global_metadata_copy = deepcopy(global_metadata)

    return {**global_metadata_copy, anchor_point.root_slot: resources}


def lookup_slot_in_resource(
    *, resource: Json, slot_name: str
) -> Union[Json, list[Json]]:
    """Lookup a slot in a resource. Raises an error if the slot does not exist."""

    content = resource.get(slot_name)

    if content is None:
        raise SlotNotFoundError(
            f"Could not find slot '{slot_name}' in resource: {resource}"
        )

    return content



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata."""

from copy import deepcopy

from metldata.builtin_transformations.merge_slots.models import SlotMergeInstruction
from metldata.custom_types import Json
from metldata.metadata_utils import (
    SlotNotFoundError,
    get_resources_of_class,
    lookup_slot_in_resource,
    upsert_resources_in_metadata,
)
from metldata.model_utils.anchors import AnchorPoint
from metldata.transform.base import MetadataTransformationError


def apply_merge_instruction_to_resource(
    *, resource: Json, merge_instruction: SlotMergeInstruction
) -> Json:
    """Merge slots in a metadata resource according to the given instruction.

    Args:
        resource: The resource to transform.
        merge_instruction: The merge instruction to apply.

    Returns:
        The transformed resource.
    """

    modified_resource = deepcopy(resource)

    try:
        lookup_slot_in_resource(
            resource=resource, slot_name=merge_instruction.target_slot
        )
    except SlotNotFoundError:
        # this is expected
        pass
    else:
        raise MetadataTransformationError(
            f"Target slot {merge_instruction.target_slot} already exists in resource"
            + f" of target class {merge_instruction.class_name}."
        )

    modified_resource[merge_instruction.target_slot] = []
    for source_slot in merge_instruction.source_slots:
        content = lookup_slot_in_resource(resource=resource, slot_name=source_slot)
        if isinstance(content, list):
            modified_resource[merge_instruction.target_slot].extend(content)
        else:
            modified_resource[merge_instruction.target_slot].append(content)

    return modified_resource


def apply_merge_instruction_to_metadata(
    *,
    metadata: Json,
    merge_instruction: SlotMergeInstruction,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Merge slots in metadata according to the given instruction.

    Args:
        metadata: The metadata to transform.
        merge_instruction: The merge instruction to apply.

    Returns:
        The transformed metadata.
    """

    resources = get_resources_of_class(
        class_name=merge_instruction.class_name,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target,
    )

    modified_resources = [
        apply_merge_instruction_to_resource(
            resource=resource, merge_instruction=merge_instruction
        )
        for resource in resources
    ]

    return upsert_resources_in_metadata(
        resources=modified_resources,
        class_name=merge_instruction.class_name,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target,
    )


def apply_merge_instructions_to_metadata(
    *,
    metadata: Json,
    merge_instructions: list[SlotMergeInstruction],
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Merge slots in metadata according to the given instructions.

    Args:
        metadata: The metadata to transform.
        merge_instructions: The merge instructions to apply.

    Returns:
        The transformed metadata.
    """

    for merge_instruction in merge_instructions:
        metadata = apply_merge_instruction_to_metadata(
            metadata=metadata,
            merge_instruction=merge_instruction,
            anchor_points_by_target=anchor_points_by_target,
        )

    return metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata models."""


from copy import deepcopy
from typing import Optional, cast

from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition

from metldata.model_utils.anchors import get_anchors_points_by_target
from metldata.model_utils.essentials import ExportableSchemaView, MetadataModel
from metldata.model_utils.identifiers import get_class_identifiers
from metldata.model_utils.manipulate import upsert_class
from metldata.transform.base import MetadataModelTransformationError


def unset_identifier(
    *, class_definition: ClassDefinition, class_identifiers: dict[str, Optional[str]]
) -> ClassDefinition:
    """Get a modified copy of the provided class definition with the identifier being
    unset.
    """

    identifier_slot_name = class_identifiers[class_definition.name]
    if identifier_slot_name is None:
        raise RuntimeError(  # This should never happen
            f"Class {class_definition.name} does not have an identifier."
        )

    modified_class = deepcopy(class_definition)

    if not modified_class.slot_usage:
        modified_class.slot_usage = {}
    elif not isinstance(modified_class.slot_usage, dict):
        raise RuntimeError(  # This should never happen
            f"Class {class_definition.name} has a slot_usage that is not a dict."
        )

    identifier_slot = modified_class.slot_usage.get(
        identifier_slot_name,
        SlotDefinition(name=identifier_slot_name),
    )
    identifier_slot = cast(SlotDefinition, identifier_slot)
    identifier_slot.identifier = False

    modified_class.slot_usage[identifier_slot_name] = identifier_slot

    return modified_class


def get_accession_slot(*, accession_slot_name: str) -> SlotDefinition:
    """Generate a slot definition for the accession slot."""

    return SlotDefinition(
        name=accession_slot_name,
        identifier=True,
        required=True,
    )


def add_accessions_to_class(
    *,
    class_definition: ClassDefinition,
    accession_slot_name: str,
    class_identifiers: dict[str, Optional[str]],
) -> ClassDefinition:
    """Get a modified copy of the provided class definition with the accession slot
    being added as identifier.
    """

    modified_class = unset_identifier(
        class_definition=class_definition, class_identifiers=class_identifiers
    )

    if not isinstance(modified_class.slots, list):
        raise RuntimeError(  # This should never happen
            f"Class {class_definition.name} has a slots that is not a list."
        )
    modified_class.slots.append(accession_slot_name)

    modified_class.slot_usage[accession_slot_name] = get_accession_slot(
        accession_slot_name=accession_slot_name
    )

    return modified_class


def add_global_accession_slot(
    *,
    schema_view: ExportableSchemaView,
    accession_slot_name: str,
    accession_slot_description: str,
) -> ExportableSchemaView:
    """Get a modified copy of the provided schema view with the accession slot being
    added to the global scope.
    """

    if schema_view.get_slot(slot_name=accession_slot_name):
        raise MetadataModelTransformationError(
            f"The slot name '{accession_slot_name}' cannot be used as accession, it is"
            + " already in use."
        )

    accession_slot = SlotDefinition(
        name=accession_slot_name, description=accession_slot_description
    )

    modified_schema_view = deepcopy(schema_view)
    modified_schema_view.add_slot(slot=accession_slot)

    return modified_schema_view


def add_accessions_to_model(
    *, model: MetadataModel, accession_slot_name: str, accession_slot_description: str
) -> MetadataModel:
    """Get a modified copy of the provided model with the accession slot being added as
    identifier.
    """

    anchor_points_by_target = get_anchors_points_by_target(model=model)
    class_identifiers = {
        class_name: identifier
        for class_name, identifier in get_class_identifiers(model=model).items()
        if class_name in anchor_points_by_target
    }
    schema_view = model.schema_view

    modified_schema_view = add_global_accession_slot(
        schema_view=schema_view,
        accession_slot_name=accession_slot_name,
        accession_slot_description=accession_slot_description,
    )

    for class_name in class_identifiers:
        class_definition = modified_schema_view.get_class(class_name=class_name)
        if not class_definition:
            raise RuntimeError(  # This should never happen
                f"Class with name '{class_name}' does not exist."
            )

        class_definition = add_accessions_to_class(
            class_definition=class_definition,
            accession_slot_name=accession_slot_name,
            class_identifiers=class_identifiers,
        )
        modified_schema_view = upsert_class(
            schema_view=modified_schema_view, class_definition=class_definition
        )

    return modified_schema_view.export_model()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models used to describe embedding profiles."""

from pydantic import BaseSettings, Field, validator

from metldata.builtin_transformations.custom_embeddings.embedding_profile import (
    EmbeddingProfile,
)


def _get_target_class_names(*, embedding_profile: EmbeddingProfile) -> list[str]:
    """A function to get the names of the target classes from an embedding profile
    including nested embedding profiles."""

    embedded_classes = [embedding_profile.target_class]
    for referenced_profile in embedding_profile.embedded_references.values():
        if isinstance(referenced_profile, EmbeddingProfile):
            embedded_classes.extend(
                _get_target_class_names(embedding_profile=referenced_profile)
            )

    return embedded_classes


class CustomEmbeddingConfig(BaseSettings):
    """Config to describe profiles for custom embeddings of classes from a metadata
    model."""

    embedding_profiles: list[EmbeddingProfile] = Field(
        ...,
        description=(
            "A list of custom embedding profiles for classes from a metadata model."
        ),
    )

    # pylint: disable=no-self-argument
    @validator("embedding_profiles")
    def check_embedding_profiles_unique(
        cls,
        value: list[EmbeddingProfile],
    ) -> list[EmbeddingProfile]:
        """Check that names for embedded classes are unique among the embedding_profiles."""

        embedded_classes = [
            embedded_class
            for profile in value
            for embedded_class in _get_target_class_names(embedding_profile=profile)
        ]

        if len(embedded_classes) != len(set(embedded_classes)):
            raise ValueError("Names for embedded classes must be unique.")

        return value

    class Config:
        """Pydantic config."""

        extra = "forbid"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Minimal models for LinkML elements."""

from dataclasses import dataclass, field
from typing import Iterable


@dataclass(frozen=True)
class MinimalSlot:
    """A minimal representation of a LinkML slot without a name"""

    range: str
    multivalued: bool


@dataclass(frozen=True)
class MinimalNamedSlot(MinimalSlot):
    """A minimal representation of a LinkML slot with a name."""

    slot_name: str

    def __lt__(self, other):
        return (
            self.slot_name < other.slot_name
            and self.multivalued < other.multivalued
            and self.range < other.range
        )


class MinimalClass(frozenset[MinimalNamedSlot]):
    """A minimal representation of a LinkML class"""


class NamedClassConflict(RuntimeError):
    """Raised when an attempt is made to add two classes of the same name but
    conflicting definitions to the same model."""


@dataclass
class MinimalLinkMLModel:
    """A minimal representation of a LinkML model in form of a dictionary
    mapping anonymous classes to their names."""

    __AUTO_GENERATION_PREFIX = "AutoGeneratedClass"

    named_classes: dict[str, MinimalClass] = field(default_factory=dict)
    anonymous_classes: dict[MinimalClass, str] = field(default_factory=dict)

    def all_classes(self) -> Iterable[tuple[str, MinimalClass]]:
        """Iterates all class names and definitions currently in the model."""
        for cls_name, cls_def in self.named_classes.items():
            yield cls_name, cls_def
        for cls_def, cls_name in self.anonymous_classes.items():
            yield cls_name, cls_def

    def add_named_class(self, cls_def: MinimalClass, cls_name: str) -> None:
        """Adds a named class. Raises an error if the class name is already used
        for a different class definition."""
        if (
            cls_name in self.named_classes and cls_def != self.named_classes[cls_name]
        ) or cls_name.startswith(self.__AUTO_GENERATION_PREFIX):
            raise NamedClassConflict(cls_name)
        self.named_classes[cls_name] = cls_def

    def add_anonymous_class(self, cls_def: MinimalClass) -> str:
        """Adds an anonymous class and returns a generated class name. If the
        same class was previously defined, the class is re-used."""
        if cls_def in self.anonymous_classes:
            cls_name = self.anonymous_classes[cls_def]
        else:
            cls_name = f"{self.__AUTO_GENERATION_PREFIX}{len(self.anonymous_classes)}"
            self.anonymous_classes[cls_def] = cls_name
        return cls_name



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config parameters and their defaults."""

from pydantic import BaseSettings, Field


class SlotDeletionConfig(BaseSettings):
    """Config containing slots to be deleted from models and associated metadata."""

    slots_to_delete: dict[str, list[str]] = Field(
        ...,
        description=(
            "A nested dictionary specifying slots that should be deleted per class."
            + " The keys refer to classes, the values to the slots that should be"
            + " deleted from the respective class."
        ),
        example={
            "class_a": ["some_slot", "another_slot"],
            "class_b": ["some_slot"],
            "class_c": ["some_slot", "yet_another_slot"],
        },
    )

    class Config:
        """Pydantic config."""

        extra = "forbid"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Generate global summary statistics."""

from typing import Any, Optional, cast

from ghga_service_commons.utils.utc_dates import now_as_utc

from metldata.artifacts_rest.models import (
    ArtifactInfo,
    GlobalStats,
    ResourceStats,
    ValueCount,
)
from metldata.load.aggregator import DbAggregator

STATS_COLLECTION_NAME = "stats"


def get_stat_slot(resource_class: str) -> Optional[str]:
    """Get the name of the slot that shall be used as grouping key."""
    if resource_class.endswith("File"):
        return "format"
    if resource_class.endswith("Protocol"):
        return "type"
    if resource_class.endswith("Individual"):
        return "sex"
    return None


async def create_stats_using_aggregator(
    artifact_infos: dict[str, ArtifactInfo],
    primary_artifact_name: str,
    db_aggregator: DbAggregator,
) -> None:
    """Create summary by running an aggregation pipeline on the database."""
    resource_stats: dict[str, ResourceStats] = {}
    artifact_name = primary_artifact_name
    artifact_info = artifact_infos[artifact_name]
    for resource_class in artifact_info.resource_classes:
        collection_name = f"art_{artifact_name}_class_{resource_class}"

        pipeline: list[dict[str, Any]] = [{"$count": "count"}]
        result = await db_aggregator.aggregate(
            collection_name=collection_name, pipeline=pipeline
        )
        if not result:
            continue
        resource_stats[resource_class] = cast(ResourceStats, result[0])

        stat_slot = get_stat_slot(resource_class)
        if not stat_slot:
            continue

        pipeline = [{"$group": {"_id": f"$content.{stat_slot}", "count": {"$sum": 1}}}]
        result = await db_aggregator.aggregate(
            collection_name=collection_name, pipeline=pipeline
        )
        if not result:
            continue

        stats: list[ValueCount] = [
            {"value": group["_id"] or "unknown", "count": group["count"]}
            for group in result
        ]
        resource_stats[resource_class]["stats"] = {stat_slot: stats}

    if resource_stats:
        global_stats = GlobalStats(
            id="global", created=now_as_utc(), resource_stats=resource_stats
        )
        stats_dao = await db_aggregator.get_dao(
            name=STATS_COLLECTION_NAME, dto_model=GlobalStats, id_field="id"
        )
        await stats_dao.upsert(global_stats)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""This module provides the DataSubgraph class to traverse a subgraph of a
LinkML-based JSON data graph."""

from collections import defaultdict
from typing import Any, Iterator, Optional

from linkml_runtime.linkml_model import SlotDefinition

from metldata.builtin_transformations.aggregate.cached_model import CachedMetadataModel
from metldata.custom_types import Json
from metldata.metadata_utils import get_resource_dict_of_class
from metldata.model_utils.anchors import AnchorPoint
from metldata.model_utils.essentials import MetadataModel


class DataTraversalError(RuntimeError):
    """Error raised when JSON data cannot be traversed as expected."""


def _resolve_path(
    *, model: MetadataModel, all_classes: list[str], origin: str, slot_names: list[str]
) -> list[SlotDefinition]:
    cur_cls = origin
    resolved_path = []
    for slot_name in slot_names:
        if cur_cls is None:
            raise DataTraversalError(
                f"Unable to resolve path '{slot_names}': Cannot traverse beyond"
                f" '{slot_names[len(resolved_path)]}', range is type or enum."
            )
        try:
            slot_def = model.schema_view.induced_slot(slot_name, cur_cls)
        except ValueError as error:
            raise DataTraversalError(
                f"Unable to find slot '{slot_name}' for class '{cur_cls}'."
            ) from error
        resolved_path.append(slot_def)
        cur_cls = slot_def.range if slot_def.range in all_classes else None
    return resolved_path


class DataSubgraph:
    """
    Given that LinkML models enable references between objects, any JSON data
    structured according to a LinkML model can be treated as a connected graph
    of objects, i.e. a data graph. A DataSubgraph represents a subgraph of the
    object graph described by a set of paths in the entity relationship model of
    the LinkML schema. Note that a path in the entity relationship model
    corresponds to a subgraph in the data graph, since the entity relationships
    can have various cardinalities (i.e. LinkML slots may be multivalued).

    The DataSubgraph enables to iterate through the nodes of the subgraph.
    """

    _model: MetadataModel
    _paths: list[list[SlotDefinition]]
    _class_identifiers: dict[str, Optional[str]]
    _all_classes: list[str]
    _anchor_points: dict[str, AnchorPoint]

    def _get_class_identifier(self, class_name: str) -> str:
        """Returns the identifier slot name for the given class name.

        Args:
            class_name (str): Name of the class

        Raises:
            DataTraversalError: If the class is unknown or does not have an
            identifier slot

        Returns:
            str: The name of the identifier slot.
        """
        slot_name = self._class_identifiers[class_name]
        if slot_name is None:
            raise DataTraversalError(
                f"Identifier slot for class '{class_name}' not found."
            )
        return slot_name

    def _resolve_non_inlined(
        self, *, identifiers: list[Any], class_name: str
    ) -> list[Json]:
        """Resolves a list of identifiers to the corresponding objects based on
        the provided class name.

        Args:
            identifiers (list[Any]): A list of identifiers
            class_name (str): The name of the corresponding class

        Raises:
            DataTraversalError: Raised when an identifier cannot be resolved.

        Returns:
            list[Json]: A list of objects corresponding to the identifiers.
        """
        try:
            return [
                self._resources_by_id[class_name][next_node]
                for next_node in identifiers
            ]
        except KeyError as error:
            raise DataTraversalError(
                f"Unable to resolve ID '{error.args[0]}' for class '{class_name}'"
            ) from error

    def terminal_nodes(self, data: Json) -> Iterator[Any]:  # noqa: C901
        """Returns a generator for all data nodes corresponding to the model
        path leaves.

        Args:
            data (Json): The data

        Yields:
            Iterator[Any]: The generator.
        """
        # Inner nodes that are blocked from being revisited. Nodes are
        # represented as (class, id) tuples. Non-identifiable classes cannot be
        # prevented from being re-visited.
        do_not_revisit: set[tuple[str, Any]] = set()
        # The stack that guides the traversal
        for path in self._paths:
            stack: list[tuple[int, Json]] = [(0, data)]
            while stack:
                depth, node = stack.pop()
                # Yield if we're at the end of the path
                if depth == len(path):
                    yield node
                    continue
                # Otherwise, add intermediate nodes to the stack unless they are None
                slot_def = path[depth]
                try:
                    next_nodes = node[slot_def.name]
                except KeyError:
                    if not slot_def.required:
                        continue
                    raise
                if not slot_def.multivalued:
                    next_nodes = [next_nodes]
                # Resolve non-inlined nodes
                if slot_def.range in self._all_classes and not slot_def.inlined:
                    next_nodes = self._resolve_non_inlined(
                        identifiers=next_nodes, class_name=slot_def.range
                    )
                # Add elements to stack
                next_range = path[depth].range
                if next_range in self._visit_once_classes:
                    for next_node in next_nodes:
                        next_hash = (
                            next_range,
                            next_node[self._get_class_identifier(next_range)],
                        )
                        if next_hash not in do_not_revisit:
                            stack.append((depth + 1, next_node))
                            do_not_revisit.add(next_hash)
                else:
                    stack.extend((depth + 1, next_node) for next_node in next_nodes)

    def _map_resources_by_id(self, submission_data: Json) -> dict[str, dict[str, Json]]:
        """Creates a mapping from class names to mappings from class identifiers
        to objects.

        Args:
            submission_data (Json): The submission data

        Returns:
            dict[str, dict[str, Json]]: The resulting mapping
        """
        resources_by_id: defaultdict[str, dict[str, Json]] = defaultdict(dict)
        for path in self._paths:
            for slot_def in path[:-1]:
                if not slot_def.inlined and slot_def.range not in resources_by_id:
                    if (
                        slot_def.range is None
                        or slot_def.range not in self._all_classes
                    ):
                        raise DataTraversalError(
                            f"Intermediate path slot '{slot_def.name}' does"
                            " not have a class range."
                        )
                    resources_by_id[slot_def.range] = get_resource_dict_of_class(
                        class_name=slot_def.range,
                        global_metadata=submission_data,
                        anchor_points_by_target=self._anchor_points,
                    )
            slot_def = path[-1]
            if (
                slot_def.range in self._all_classes
                and not slot_def.inlined
                and slot_def.range not in resources_by_id
            ):
                resources_by_id[slot_def.range] = get_resource_dict_of_class(
                    class_name=slot_def.range,
                    global_metadata=submission_data,
                    anchor_points_by_target=self._anchor_points,
                )

        return resources_by_id

    def __init__(
        self,
        *,
        model: CachedMetadataModel,
        submission_data: Json,
        origin: str,
        path_strings: list[str],
        visit_once_classes: Optional[list[str]] = None,
    ):
        """Creates a new DataSubgraph object.

        Args:
            model (MetadataModel): The LinkML metadata model
            submission_data (Json): The full submission data, a representation
            of the LinkML tree root class
            model_paths (list[ModelPath]): A list of model paths
            visit_once_classes (Optional[list[str]], optional): List of classes
            for which objects shall be traversed only once. Defaults to None.
        """
        self._model = model.model
        self._anchor_points = model.anchors_points_by_target
        self._visit_once_classes = visit_once_classes if visit_once_classes else []
        self._all_classes = list(self._model.schema_view.all_classes())
        self._paths = [
            _resolve_path(
                model=self._model,
                origin=origin,
                slot_names=path_string.split("."),
                all_classes=self._all_classes,
            )
            for path_string in path_strings
        ]
        self._resources_by_id = self._map_resources_by_id(submission_data)
        self._class_identifiers = {
            cls_name: ap.identifier_slot for cls_name, ap in self._anchor_points.items()
        }



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Joint fixture with setup for event content"""

import json
from dataclasses import dataclass
from typing import Any

from ghga_service_commons.api.testing import AsyncTestClient
from hexkit.providers.akafka.testutils import KafkaFixture
from pytest_asyncio import fixture

from metldata.artifacts_rest.artifact_info import ArtifactInfo, load_artifact_info
from metldata.custom_types import Json
from metldata.load.auth import generate_token_and_hash
from metldata.load.config import ArtifactLoaderAPIConfig
from metldata.load.main import get_app
from metldata.load.models import ArtifactResourceDict
from metldata.model_utils.essentials import MetadataModel
from tests.fixtures.load.config import get_config
from tests.fixtures.load.utils import BASE_DIR
from tests.fixtures.mongodb import (  # noqa: F401; pylint: disable=unused-import
    MongoDbFixture,
    mongodb_fixture,
)

ARTIFACT_INFOS_PATH = BASE_DIR / "example_models" / "artifact_infos.json"
ARTIFACT_NAME = "embedded_public"
EMBEDDED_ARTIFACT_PATH = BASE_DIR / "example_metadata" / f"{ARTIFACT_NAME}.json"
EMBEDDED_DATASET_TEST_PATH = BASE_DIR / "example_metadata" / "embedded_dataset.json"
EMBEDDED_ARTIFACT_MODEL_PATH = (
    BASE_DIR / "example_models" / f"{ARTIFACT_NAME}_model.yaml"
)


@dataclass
class JointFixture:
    artifact_infos: list[ArtifactInfo]
    artifact_resources: dict[str, Any]
    client: AsyncTestClient
    config: ArtifactLoaderAPIConfig
    expected_file_resource_content: Json
    expected_embedded_dataset_resource_content: Json
    kafka: KafkaFixture
    mongodb: MongoDbFixture
    token: str


@fixture
async def joint_fixture(
    kafka_fixture: KafkaFixture,
    mongodb_fixture: MongoDbFixture,  # noqa: F811
) -> JointFixture:
    """Get a tuple of a configured test client together with a corresponding token."""

    artifact_infos = [
        load_artifact_info(
            name=ARTIFACT_NAME,
            description=ARTIFACT_NAME,
            model=MetadataModel.init_from_path(EMBEDDED_ARTIFACT_MODEL_PATH),
        )
    ]

    with open(EMBEDDED_ARTIFACT_PATH, "r", encoding="utf-8") as file:
        raw_artifacts = json.load(file)
        artifacts: ArtifactResourceDict = {
            raw_artifacts["type_"]: [raw_artifacts["payload"]["content"]]
        }

    token, token_hash = generate_token_and_hash()

    config = get_config(
        [
            {
                "artifact_infos": artifact_infos,
                "loader_token_hashes": [token_hash],
            },
            kafka_fixture.config,
            mongodb_fixture.config,
        ]
    )

    expected_file_resource_content = artifacts["embedded_public"][0]["study_files"][0]
    expected_embedded_dataset_resource_content = artifacts["embedded_public"][0][
        "embedded_dataset"
    ][1]

    app = await get_app(config=config)
    client = AsyncTestClient(app)
    return JointFixture(
        artifact_infos=artifact_infos,
        artifact_resources=artifacts,
        client=client,
        config=config,
        expected_file_resource_content=expected_file_resource_content,
        expected_embedded_dataset_resource_content=expected_embedded_dataset_resource_content,
        kafka=kafka_fixture,
        mongodb=mongodb_fixture,
        token=token,
    )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata models."""

from linkml_runtime.linkml_model.meta import SlotDefinition

from metldata.builtin_transformations.infer_references.reference import (
    InferredReference,
)
from metldata.model_utils.essentials import MetadataModel
from metldata.model_utils.manipulate import (
    ModelManipulationError,
    add_slot_if_not_exists,
    upsert_class_slot,
)
from metldata.transform.base import MetadataModelTransformationError


def inferred_reference_to_slot(reference: InferredReference) -> SlotDefinition:
    """Convert an inferred reference into a slot definition to be mounted on the
    source class."""

    return SlotDefinition(
        name=reference.new_slot,
        range=reference.target,
        multivalued=reference.multivalued,
        inlined=False,
        required=True,
    )


def add_reference_to_model(
    *, model: MetadataModel, reference: InferredReference
) -> MetadataModel:
    """Get a modified copy of the provided model with the inferred reference being
    added.

    Raises:
            MetadataModelTransformationError:
                if the transformation of the metadata model fails.
    """

    new_slot = inferred_reference_to_slot(reference)

    schema_view = model.schema_view
    try:
        schema_view = add_slot_if_not_exists(schema_view=schema_view, new_slot=new_slot)
        schema_view = upsert_class_slot(
            schema_view=schema_view, class_name=reference.source, new_slot=new_slot
        )
    except ModelManipulationError as error:
        raise MetadataModelTransformationError(
            f"Failed to add the inferred reference '{reference}' to the metadata"
            + f" model.: {error}"
        ) from error

    return schema_view.export_model()


def add_references_to_model(
    *, model: MetadataModel, references: list[InferredReference]
) -> MetadataModel:
    """Transform the metadata model and return the transformed one.
    Raises:
            MetadataModelTransformationError:
                if the transformation of the metadata model fails.
    """

    for reference in references:
        model = add_reference_to_model(model=model, reference=reference)

    return model



#!/usr/bin/env python3

# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This script evaluates the entries in .static_files, .mandatory_files and
.deprecated_files and compares them with the microservice template repository,
or verifies their existence or non-existence depending on the list they are in.
"""

import difflib
import os
import shutil
import stat
import sys
import urllib.error
import urllib.parse
import urllib.request
from pathlib import Path

try:
    from script_utils.cli import echo_failure, echo_success, run
except ImportError:
    echo_failure = echo_success = print

    def run(main_fn):
        """Run main function without cli tools (typer)."""
        main_fn(check="--check" in sys.argv[1:])


REPO_ROOT_DIR = Path(__file__).parent.parent.resolve()

DEPRECATED_FILES = ".deprecated_files"
MANDATORY_FILES = ".mandatory_files"
STATIC_FILES = ".static_files"
IGNORE_SUFFIX = "_ignore"
RAW_TEMPLATE_URL = (
    "https://raw.githubusercontent.com/ghga-de/microservice-repository-template/main/"
)


class ValidationError(RuntimeError):
    """Raised when files need to be updated."""


def get_file_list(list_name: str) -> list[str]:
    """Return a list of all file names specified in a given list file."""
    list_path = REPO_ROOT_DIR / list_name
    with open(list_path, "r", encoding="utf8") as list_file:
        file_list = [
            clean_line
            for clean_line in (
                line.rstrip() for line in list_file if not line.startswith("#")
            )
            if clean_line
        ]
    if not list_name.endswith(IGNORE_SUFFIX):
        ignore_list_name = list_name + IGNORE_SUFFIX
        try:
            file_set_ignore = set(get_file_list(ignore_list_name))
        except FileNotFoundError:
            print(f"  - {ignore_list_name} is missing, no exceptions from the template")
        else:
            file_list = [line for line in file_list if line not in file_set_ignore]
    return file_list


def get_template_file_content(relative_file_path: str):
    """Get the content of the template file corresponding to the given path."""
    remote_file_url = urllib.parse.urljoin(RAW_TEMPLATE_URL, relative_file_path)
    remote_file_request = urllib.request.Request(remote_file_url)
    try:
        with urllib.request.urlopen(remote_file_request) as remote_file_response:
            return remote_file_response.read().decode(
                remote_file_response.headers.get_content_charset("utf-8")
            )
    except urllib.error.HTTPError as remote_file_error:
        print(
            f"  - WARNING: request to remote file {remote_file_url} returned"
            f" status code {remote_file_error.code}"
        )
        return None


def diff_content(local_file_path, local_file_content, template_file_content) -> bool:
    """Show diff between given local and remote template file content."""
    if local_file_content != template_file_content:
        print(f"  - {local_file_path}: differs from template")
        for line in difflib.unified_diff(
            template_file_content.splitlines(keepends=True),
            local_file_content.splitlines(keepends=True),
            fromfile="template",
            tofile="local",
        ):
            print("   ", line.rstrip())
        return True
    return False


def check_file(relative_file_path: str, diff: bool = False) -> bool:
    """Compare file at the given path with the given content.

    Returns True if there are differences.
    """
    local_file_path = REPO_ROOT_DIR / Path(relative_file_path)

    if not local_file_path.exists():
        print(f"  - {local_file_path} does not exist")
        return True

    if diff:
        template_file_content = get_template_file_content(relative_file_path)

        if template_file_content is None:
            print(f"  - {local_file_path}: cannot check, remote is missing")
            return True

        with open(local_file_path, "r", encoding="utf8") as file:
            return diff_content(local_file_path, file.read(), template_file_content)

    return False


def update_file(relative_file_path: str, diff: bool = False) -> bool:
    """Update file at the given relative path.

    Returns True if there are updates.
    """

    local_file_path = REPO_ROOT_DIR / Path(relative_file_path)
    local_parent_dir = local_file_path.parent

    if not local_parent_dir.exists():
        local_parent_dir.mkdir(parents=True)

    if diff or not local_file_path.exists():
        template_file_content = get_template_file_content(relative_file_path)

        if template_file_content is None:
            print(f"  - {local_file_path}: cannot update, remote is missing")
            return True

        if diff and local_file_path.exists():
            with open(local_file_path, "r", encoding="utf8") as file:
                if file.read() == template_file_content:
                    return False

        executable = template_file_content.startswith("#!")
        executable_flags = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
        with open(local_file_path, "w", encoding="utf8") as file:
            file.write(template_file_content)
            mode = os.fstat(file.fileno()).st_mode
            if executable:
                mode |= executable_flags
            else:
                mode &= ~executable_flags
            os.fchmod(file.fileno(), stat.S_IMODE(mode))

        print(f"  - {local_file_path}: updated")
        return True

    return False


def update_files(files: list[str], check: bool = False, diff: bool = False) -> bool:
    """Update or check all the files in the given list.

    Returns True if there are updates.
    """
    updates = False
    update_or_check_file = check_file if check else update_file
    for relative_file_path in files:
        if update_or_check_file(relative_file_path, diff=diff):
            updates = True
    return updates


def remove_files(files: list[str], check: bool = False) -> bool:
    """Remove or check all the files in the given list.

    Returns True if there are updates.
    """
    updates = False
    for relative_file_path in files:
        local_file_path = REPO_ROOT_DIR / Path(relative_file_path)

        if local_file_path.exists():
            if check:
                print(f"  - {local_file_path}: deprecated, but exists")
            else:
                if local_file_path.is_dir():
                    shutil.rmtree(local_file_path)
                else:
                    local_file_path.unlink()
                print(f"  - {local_file_path}: removed, since it is deprecated")
            updates = True
    return updates


def main(check: bool = False):
    """Update the static files in the service template."""
    updated = False
    if not check:
        update_files([STATIC_FILES], diff=True, check=False)

    print("Static files...")
    files_to_update = get_file_list(STATIC_FILES)
    if check:
        files_to_update.append(STATIC_FILES)
    files_to_update.extend((MANDATORY_FILES, DEPRECATED_FILES))
    if update_files(files_to_update, diff=True, check=check):
        updated = True

    print("Mandatory files...")
    files_to_guarantee = get_file_list(MANDATORY_FILES)
    if update_files(files_to_guarantee, check=check):
        updated = True

    print("Deprecated files...")
    files_to_remove = get_file_list(DEPRECATED_FILES)
    if remove_files(files_to_remove, check=check):
        updated = True

    if check:
        if updated:
            echo_failure("Validating files from template failed.")
            sys.exit(1)
        echo_success("Successfully validated files from template.")
    else:
        echo_success(
            "Successfully updated files from template."
            if updated
            else "No updates from the template were necessary."
        )


if __name__ == "__main__":
    run(main)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Data models."""

from typing import Optional, TypedDict

from ghga_service_commons.utils.utc_dates import DateTimeUTC
from pydantic import BaseModel, Field, validator

from metldata.custom_types import Json
from metldata.model_utils.anchors import AnchorPoint

try:  # workaround for https://github.com/pydantic/pydantic/issues/5821
    from typing_extensions import Literal
except ImportError:
    from typing import Literal  # type: ignore


class ArtifactResource(BaseModel):
    """Information on a resource of an artifact."""

    id_: str = Field(..., description="The ID of the resource.")
    class_name: str = Field(
        ..., description="The name of the class this resource corresponds to."
    )
    content: Json = Field(..., description="The metadata content of that resource.")


class ArtifactResourceClass(BaseModel):
    """Model to describe a resource class of an artifact."""

    name: str = Field(..., description="The name of the metadata class.")
    description: Optional[str] = Field(
        None, description="A description of the metadata class."
    )
    anchor_point: AnchorPoint = Field(
        ..., description="The anchor point for this metadata class."
    )
    json_schema: Json = Field(
        ..., description="The JSON schema for this metadata class."
    )


class ArtifactInfo(BaseModel):
    """Model to describe general information on an artifact.
    Please note, it does not contain actual artifact instances derived from specific
    metadata.
    """

    name: str = Field(..., description="The name of the artifact.")
    description: str = Field(..., description="A description of the artifact.")

    resource_classes: dict[str, ArtifactResourceClass] = Field(
        ...,
        description=(
            "A dictionary of resource classes for this artifact."
            + " The keys are the names of the classes."
            + " The values are the corresponding class models."
        ),
    )

    # pylint: disable=no-self-argument
    @validator("resource_classes")
    def check_resource_class_names(
        cls, value: dict[str, ArtifactResourceClass]
    ) -> dict[str, ArtifactResourceClass]:
        """Check if the keys of the `resource_classes` dictionary correspond to the
        names of the metadata classes.
        """

        for class_name, resource_class in value.items():
            if class_name != resource_class.name:
                raise ValueError(
                    f"Key '{class_name}' of 'resource_classes' does not match"
                    + f" the name '{resource_class.name}' of the corresponding"
                    + " metadata class."
                )

        return value


class ResourceCount(TypedDict):
    """Number of instances of a resource."""

    count: int


class ValueCount(ResourceCount):
    """Number of instances of a certain value."""

    value: str


class ResourceStats(ResourceCount, total=False):
    """Summary statistics for a resource."""

    stats: dict[str, list[ValueCount]]


class GlobalStats(BaseModel):
    """Model to describe statistical information on all resources."""

    id: Literal["global"]
    created: DateTimeUTC = Field(..., description="When these stats were created.")

    resource_stats: dict[str, ResourceStats] = Field(
        ...,
        description=(
            "A dictionary of global resource stats."
            + " The keys are the names of the classes."
            + " The values are the corresponding summary statistics."
        ),
    )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Valid and invalid metadata examples using the minimal model."""

from typing import Any

import yaml

from tests.fixtures.utils import BASE_DIR

EXAMPLE_METADATA_DIR = BASE_DIR / "example_metadata"


def _get_example_metadata(name: str) -> dict[str, Any]:
    """Get example metadata."""

    with open(EXAMPLE_METADATA_DIR / f"{name}.yaml") as file:
        return yaml.safe_load(file)


VALID_MINIMAL_METADATA_EXAMPLES = [
    _get_example_metadata(f"minimal_metadata_{idx}") for idx in range(1, 3)
]
VALID_MINIMAL_METADATA_EXAMPLE = VALID_MINIMAL_METADATA_EXAMPLES[0]
INVALID_MINIMAL_METADATA_EXAMPLES = [
    _get_example_metadata(f"minimal_metadata_{invalid_example}")
    for invalid_example in [
        "additional_field",
        "missing_field",
        "invalid_enum",
        "invalid_int",
    ]
]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Event publisher for population and deletion events to MASS, WPS and claims repository"""

import json
from abc import ABC, abstractmethod

from ghga_event_schemas.pydantic_ import (
    MetadataDatasetID,
    MetadataDatasetOverview,
    SearchableResource,
    SearchableResourceInfo,
)
from hexkit.protocols.eventpub import EventPublisherProtocol
from pydantic import BaseSettings, Field


class EventPubTranslatorConfig(BaseSettings):
    """Config for publishing population/deletion events to other services"""

    primary_artifact_name: str = Field(
        ...,
        description="Name of the artifact from which the information for outgoing"
        + " change events is derived.",
        example="embedded_public",
    )
    primary_dataset_name: str = Field(
        ...,
        description="Name of the resource class corresponding to the embedded_dataset slot.",
        example="EmbeddedDataset",
    )

    resource_change_event_topic: str = Field(
        ...,
        description="Name of the topic used for events informing other services about"
        + " resource changes, i.e. deletion or insertion.",
        example="searchable_resources",
    )
    resource_deletion_event_type: str = Field(
        ...,
        description="Type used for events indicating the deletion of a previously"
        + " existing resource.",
        example="searchable_resource_deleted",
    )
    resource_upsertion_type: str = Field(
        ...,
        description="Type used for events indicating the upsert of a resource.",
        example="searchable_resource_upserted",
    )

    dataset_change_event_topic: str = Field(
        ...,
        description="Name of the topic announcing, among other things, the list of"
        + " files included in a new dataset.",
        example="metadata_datasets",
    )
    dataset_deletion_type: str = Field(
        ...,
        description="Type used for events announcing a new dataset overview.",
        example="dataset_deleted",
    )
    dataset_upsertion_type: str = Field(
        ...,
        description="Type used for events announcing a new dataset overview.",
        example="dataset_created",
    )


class EventPublisherPort(ABC):
    """A port through which events are communicated with the outside."""

    @abstractmethod
    async def process_dataset_deletion(self, *, accession: str):
        """Communicate the deletion of an embedded dataset resource"""

    @abstractmethod
    async def process_resource_deletion(self, *, accession: str, class_name: str):
        """Communicate the deletion of an artifact resource"""

    @abstractmethod
    async def process_dataset_upsert(
        self, *, dataset_overview: MetadataDatasetOverview
    ):
        """Communicate the upsert of an embedded dataset resource"""

    @abstractmethod
    async def process_resource_upsert(self, *, resource: SearchableResource):
        """Communicate the upsert of an artifact resource"""

    @abstractmethod
    def is_primary_dataset_source(
        self, *, artifact_name: str, resource_class_name: str
    ) -> bool:
        """Checks if combination of artifact name and resource class name describe the
        configured source for outbound change events"""


class EventPubTranslator(EventPublisherPort):
    """A translator according to  the triple hexagonal architecture implementing
    the EventPublisherPort."""

    def __init__(
        self, *, config: EventPubTranslatorConfig, provider: EventPublisherProtocol
    ):
        """Initialize with config and a provider of the EventPublisherProtocol."""

        self._config = config
        self._provider = provider

    async def process_dataset_deletion(self, *, accession: str):
        """Communicate the deletion of an embedded dataset resource

        Fires an event that should be processed by the claims repository"""

        dataset_id = MetadataDatasetID(accession=accession)

        payload = json.loads(dataset_id.json())
        await self._provider.publish(
            payload=payload,
            type_=self._config.dataset_deletion_type,
            key=f"dataset_embedded_{dataset_id.accession}",
            topic=self._config.dataset_change_event_topic,
        )

    async def process_resource_deletion(self, *, accession: str, class_name: str):
        """Communicate the deletion an artifact resource

        Fires an event that should be processed by MASS"""

        resource_info = SearchableResourceInfo(
            accession=accession, class_name=class_name
        )

        payload = json.loads(resource_info.json())
        await self._provider.publish(
            payload=payload,
            type_=self._config.resource_deletion_event_type,
            key=f"dataset_embedded_{resource_info.accession}",
            topic=self._config.resource_change_event_topic,
        )

    async def process_dataset_upsert(
        self, *, dataset_overview: MetadataDatasetOverview
    ):
        """Communicate the upsert of an embedded dataset resource

        Fires an event that should be processed by the WPS"""

        payload = json.loads(dataset_overview.json())
        await self._provider.publish(
            payload=payload,
            type_=self._config.dataset_upsertion_type,
            key=f"dataset_embedded_{dataset_overview.accession}",
            topic=self._config.dataset_change_event_topic,
        )

    async def process_resource_upsert(self, *, resource: SearchableResource):
        """Communicate the upsert of an artifact resource

        Fires an event that should be processed by MASS"""

        payload = json.loads(resource.json())
        await self._provider.publish(
            payload=payload,
            type_=self._config.resource_upsertion_type,
            key=f"dataset_embedded_{resource.accession}",
            topic=self._config.resource_change_event_topic,
        )

    def is_primary_dataset_source(
        self, *, artifact_name: str, resource_class_name: str
    ) -> bool:
        """Checks if combination of artifact name and resource class name describe the
        configured source for outbound dataset change events"""
        return (
            self._config.primary_artifact_name == artifact_name
            and self._config.primary_dataset_name == resource_class_name
        )



#!/usr/bin/env python3

# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

"""This script checks that the license and license headers
exists and that they are up to date.
"""

import argparse
import re
import sys
from datetime import date
from pathlib import Path
from typing import List, Optional, Tuple, Union

# root directory of the package:
ROOT_DIR = Path(__file__).parent.parent.resolve()

# file containing the default global copyright notice:
GLOBAL_COPYRIGHT_FILE_PATH = ROOT_DIR / ".devcontainer" / "license_header.txt"

# exclude files and dirs from license header check:
EXCLUDE = [
    ".devcontainer",
    "eggs",
    ".eggs",
    "dist",
    "build",
    "develop-eggs",
    "lib",
    "lib62",
    "parts",
    "sdist",
    "wheels",
    "pip-wheel-metadata",
    ".git",
    ".github",
    ".flake8",
    ".gitignore",
    ".pylintrc",
    "example_config.yaml",
    "config_schema.json",
    "LICENSE",  # is checked but not for the license header
    ".pre-commit-config.yaml",
    "docs",
    ".vscode",
    ".mypy_cache",
    ".mypy.ini",
    ".pytest_cache",
    ".editorconfig",
    ".static_files",
    ".static_files_ignore",
    ".mandatory_files",
    ".mandatory_files_ignore",
    ".deprecated_files",
    ".deprecated_files_ignore",
]

# exclude file by file ending from license header check:
EXCLUDE_ENDINGS = [
    "html",
    "ini",
    "jinja",
    "json",
    "md",
    "pub",
    "pyc",
    "sec",
    "txt",
    "xml",
    "yaml",
    "yml",
]

# exclude any files with names that match any of the following regex:
EXCLUDE_PATTERN = [r".*\.egg-info.*", r".*__cache__.*", r".*\.git.*"]

# The License header, "{year}" will be replaced by current year:
COPYRIGHT_TEMPLATE = """Copyright {year} {author}

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

# A list of all chars that may be used to introduce a comment:
COMMENT_CHARS = ["#"]

AUTHOR = """Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
for the German Human Genome-Phenome Archive (GHGA)"""

# The copyright notice should not date earlier than this year:
MIN_YEAR = 2021

# The path to the License file relative to target dir
LICENSE_FILE = "LICENSE"


class GlobalCopyrightNotice:
    """
    This is used to store the copyright notice that should be identical for all checked
    files.
    The text of the copyright notice is stored in the `text`
    property. This property can only be set once.
    The property `n_lines` gives the number of lines of the text. It is inferred once
    `text` is set.
    """

    def __init__(self):
        self._text: Optional[str] = None
        self._n_lines: Optional[int] = None

    @property
    def text(self) -> Optional[str]:
        return self._text

    @text.setter
    def text(self, new_text: str):
        if self._text is not None:
            raise RuntimeError("You can only set the value once.")
        self._text = new_text
        self._n_lines = len(self._text.split("\n"))

    @property
    def n_lines(self) -> int:
        if self._n_lines is None:
            raise ValueError(
                "This property is not yet available."
                + " Please set the `text` property first."
            )
        return self._n_lines


class UnexpectedBinaryFileError(RuntimeError):
    """Thrown when trying to read a binary file."""

    def __init__(self, file_path: Union[str, Path]):
        message = f"The file could not be read because it is binary: {str(file_path)}"
        super().__init__(message)


def get_target_files(
    target_dir: Path,
    exclude: List[str] = EXCLUDE,
    exclude_endings: List[str] = EXCLUDE_ENDINGS,
    exclude_pattern: List[str] = EXCLUDE_PATTERN,
) -> List[Path]:
    """Get target files that are not match the exclude conditions.
    Args:
        target_dir (pathlib.Path): The target dir to search.
        exclude (List[str], optional):
            Overwrite default list of file/dir paths relative to
            the target dir that shall be excluded.
        exclude_endings (List[str], optional):
            Overwrite default list of file endings that shall
            be excluded.
        exclude_pattern (List[str], optional):
            Overwrite default list of regex patterns match file path
            for exclusion.
    """
    abs_target_dir = Path(target_dir).absolute()
    exclude_normalized = [(abs_target_dir / excl).absolute() for excl in exclude]

    # get all files:
    all_files = [
        file_.absolute() for file_ in Path(abs_target_dir).rglob("*") if file_.is_file()
    ]

    target_files = [
        file_
        for file_ in all_files
        if not (
            any([file_.is_relative_to(excl) for excl in exclude_normalized])
            or any([str(file_).endswith(ending) for ending in exclude_endings])
            or any([re.match(pattern, str(file_)) for pattern in exclude_pattern])
        )
    ]
    return target_files


def normalized_line(line: str, chars_to_trim: List[str] = COMMENT_CHARS) -> str:
    norm_line = line.strip()

    for char in chars_to_trim:
        norm_line = norm_line.strip(char)

    return norm_line.strip("\n").strip("\t").strip()


def normalized_text(text: str, chars_to_trim: List[str] = COMMENT_CHARS) -> str:
    "Normalize a license header text."
    lines = text.split("\n")

    norm_lines: List[str] = []

    for line in lines:
        stripped_line = line.strip()
        # exclude shebang:
        if stripped_line.startswith("#!"):
            continue

        norm_line = normalized_line(stripped_line)

        # exclude empty lines:
        if norm_line == "":
            continue

        norm_lines.append(norm_line)

    return "\n".join(norm_lines).strip("\n")


def format_copyright_template(copyright_template: str, author: str) -> str:
    """Formats license header by inserting the specified author for every occurence of
    "{author}" in the header template.
    """
    return normalized_text(copyright_template.replace("{author}", author))


def is_commented_line(line: str, comment_chars: List[str] = COMMENT_CHARS) -> bool:
    """Checks whether a line is a comment."""
    line_stripped = line.strip()
    for comment_char in comment_chars:
        if line_stripped.startswith(comment_char):
            return True

    return False


def is_empty_line(line: str) -> bool:
    """Checks whether a line is empty."""
    return line.strip("\n").strip("\t").strip() == ""


def get_header(file_path: Path, comment_chars: List[str] = COMMENT_CHARS):
    """Extracts the header from a file and normalizes it."""
    header_lines: List[str] = []

    try:
        with open(file_path, "r") as file:
            for line in file:
                if is_commented_line(
                    line, comment_chars=comment_chars
                ) or is_empty_line(line):
                    header_lines.append(line)
                else:
                    break
    except UnicodeDecodeError as error:
        raise UnexpectedBinaryFileError(file_path=file_path) from error

    # normalize the lines:
    header = "".join(header_lines)
    return normalized_text(header, chars_to_trim=comment_chars)


def validate_year_string(year_string: str, min_year: int = MIN_YEAR) -> bool:
    """Check if the specified year string is valid.
    Returns `True` if valid or `False` otherwise."""

    current_year = date.today().year

    # If the year_string is a single number, it must be the current year:
    if year_string.isnumeric():
        return int(year_string) == current_year

    # Otherwise, a range (e.g. 2021 - 2023) is expected:
    match = re.match("(\d+) - (\d+)", year_string)

    if not match:
        return False

    year_1 = int(match.group(1))
    year_2 = int(match.group(2))

    # Check the validity of the range:
    if year_1 >= min_year and year_2 <= year_1:
        return False

    # year_2 must be equal to the current year:
    return year_2 == current_year


def check_copyright_notice(
    copyright: str,
    global_copyright: GlobalCopyrightNotice,
    copyright_template: str = COPYRIGHT_TEMPLATE,
    author: str = AUTHOR,
    comment_chars: List[str] = COMMENT_CHARS,
    min_year: int = MIN_YEAR,
) -> bool:
    """Checks the specified copyright text against a template.

    copyright_template (str):
        A string containing the copyright text to check against the template.
    global_copyright (str, None):
        If this is a string, it is checked whether the copyright notice in this file
        contains the same year string.
        If this is None, the variable is set to the year string present in the
        copyright notice of this file.
    copyright_template (str, optional):
        A string containing a template for the expected license header.
        You may include "{year}" which will be replace by the current year.
        This defaults to the Apache 2.0 Copyright notice.
    author (str, optional):
        The author that shall be included in the license header.
        It will replace any appearance of "{author}" in the license
        header. This defaults to an auther info for GHGA.

    """
    # If the global_copyright is already set, check if the current copyright is
    # identical to it:
    copyright_lines = copyright.split("\n")
    if global_copyright.text is not None:
        copyright_cleaned = "\n".join(copyright_lines[0 : global_copyright.n_lines])
        return global_copyright.text == copyright_cleaned

    formatted_template = format_copyright_template(copyright_template, author=author)
    template_lines = formatted_template.split("\n")

    # The header should be at least as long as the template:
    if len(copyright_lines) < len(template_lines):
        return False

    for idx, template_line in enumerate(template_lines):
        header_line = copyright_lines[idx]

        if "{year}" in template_line:
            pattern = template_line.replace("{year}", r"(.+?)")
            match = re.match(pattern, header_line)

            if not match:
                return False

            year_string = match.group(1)
            if not validate_year_string(year_string, min_year=min_year):
                return False

        elif template_line != header_line:
            return False

    # Take this copyright as the global_copyright from now on:
    copyright_cleaned = "\n".join(copyright_lines[0 : len(template_line)])
    global_copyright.text = copyright_cleaned

    return True


def check_file_headers(
    target_dir: Path,
    global_copyright: GlobalCopyrightNotice,
    copyright_template: str = COPYRIGHT_TEMPLATE,
    author: str = AUTHOR,
    exclude: List[str] = EXCLUDE,
    exclude_endings: List[str] = EXCLUDE_ENDINGS,
    exclude_pattern: List[str] = EXCLUDE_PATTERN,
    comment_chars: List[str] = COMMENT_CHARS,
    min_year: int = MIN_YEAR,
) -> Tuple[List[Path], List[Path]]:
    """Check files for presence of a license header and verify that
    the copyright notice is up to date (correct year).

    Args:
        target_dir (pathlib.Path): The target dir to search.
        copyright_template (str, optional):
            A string containing a template for the expected license header.
            You may include "{year}" which will be replace by the current year.
            This defaults to the Apache 2.0 Copyright notice.
        global_copyright (str, None):
            If this is a string, it is checked whether the copyright notice of these
            files contains the same year string.
            If this is None, the variable is set to the year string present in the
            copyright notice of these files.
        author (str, optional):
            The author that shall be included in the license header.
            It will replace any appearance of "{author}" in the license
            header. This defaults to an author info for GHGA.
        exclude (List[str], optional):
            Overwrite default list of file/dir paths relative to
            the target dir that shall be excluded.
        exclude_endings (List[str], optional):
            Overwrite default list of file endings that shall
            be excluded.
        exclude_pattern (List[str], optional):
            Overwrite default list of regex patterns match file path
            for exclusion.
    """
    target_files = get_target_files(
        target_dir,
        exclude=exclude,
        exclude_endings=exclude_endings,
        exclude_pattern=exclude_pattern,
    )

    # check if license header present in file:
    passed_files: List[Path] = []
    failed_files: List[Path] = []

    for target_file in target_files:
        try:
            header = get_header(target_file, comment_chars=comment_chars)
            if check_copyright_notice(
                copyright=header,
                global_copyright=global_copyright,
                copyright_template=copyright_template,
                author=author,
                comment_chars=comment_chars,
                min_year=min_year,
            ):
                passed_files.append(target_file)
            else:
                failed_files.append(target_file)
        except UnexpectedBinaryFileError:
            # This file is a binary and is therefor skipped.
            pass

    return (passed_files, failed_files)


def check_license_file(
    license_file: Path,
    global_copyright: GlobalCopyrightNotice,
    copyright_template: str = COPYRIGHT_TEMPLATE,
    author: str = AUTHOR,
    comment_chars: List[str] = COMMENT_CHARS,
    min_year: int = MIN_YEAR,
) -> bool:
    """Currently only checks if the copyright notice in the
    License file is up to data.

    Args:
        license_file (pathlib.Path, optional): Overwrite the default license file.
        global_copyright (str, None):
            If this is a string, it is checked whether the copyright notice in this file
            contains the same year string.
            If this is None, the variable is set to the year string present in the
            copyright notice of this file.
        copyright_template (str, optional):
            A string of the copyright notice (usually same as license header).
            You may include "{year}" which will be replace by the current year.
            This defaults to the Apache 2.0 Copyright notice.
        author (str, optional):
            The author that shall be included in the copyright notice.
            It will replace any appearance of "{author}" in the copyright
            notice. This defaults to an author info for GHGA.
    """

    if not license_file.is_file():
        print(f'Could not find license file "{str(license_file)}".')
        return False

    with open(license_file, "r") as file_:
        license_text = normalized_text(file_.read())

    # Extract the copyright notice:
    # (is expected to be at the end of the file):
    formatted_template = format_copyright_template(copyright_template, author=author)
    template_lines = formatted_template.split("\n")
    license_lines = license_text.split("\n")
    copyright = "\n".join(license_lines[-len(template_lines) :])

    return check_copyright_notice(
        copyright=copyright,
        global_copyright=global_copyright,
        copyright_template=copyright_template,
        author=author,
        comment_chars=comment_chars,
        min_year=min_year,
    )


def run():
    """Run checks from CLI."""
    parser = argparse.ArgumentParser(
        prog="license-checker",
        description=(
            "This script checks that the license and license headers "
            + "exists and that they are up to date."
        ),
    )

    parser.add_argument(
        "-L",
        "--no-license-file-check",
        help="Disables the check of the license file",
        action="store_true",
    )

    parser.add_argument(
        "-t",
        "--target-dir",
        help="Specify a custom target dir. Overwrites the default package root.",
    )

    args = parser.parse_args()

    target_dir = Path(args.target_dir).absolute() if args.target_dir else ROOT_DIR

    print(f'Working in "{target_dir}"\n')

    global_copyright = GlobalCopyrightNotice()

    # get global copyright from .devcontainer/license_header.txt file:
    with open(GLOBAL_COPYRIGHT_FILE_PATH, "r") as global_copyright_file:
        global_copyright.text = normalized_text(global_copyright_file.read())

    if args.no_license_file_check:
        license_file_valid = True
    else:
        license_file = Path(target_dir / LICENSE_FILE)
        print(f'Checking if LICENSE file is up to date: "{license_file}"')
        license_file_valid = check_license_file(
            license_file, global_copyright=global_copyright
        )
        print(
            "Copyright notice in license file is "
            + ("" if license_file_valid else "not ")
            + "up to date.\n"
        )

    print("Checking license headers in files:")
    passed_files, failed_files = check_file_headers(
        target_dir, global_copyright=global_copyright
    )
    print(f"{len(passed_files)} files passed.")
    print(f"{len(failed_files)} files failed" + (":" if failed_files else "."))
    for failed_file in failed_files:
        print(f'  - "{failed_file.relative_to(target_dir)}"')

    print("")

    if failed_files or not license_file_valid:
        print("Some checks failed.")
        sys.exit(1)

    print("All checks passed.")
    sys.exit(0)


if __name__ == "__main__":
    run()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Client functionality to drop artifacts onto the endpoint for uploading them."""

import httpx

from metldata.event_handling.event_handling import FileSystemEventCollector
from metldata.load.collect import collect_artifacts
from metldata.load.config import ArtifactLoaderClientConfig


class ArtifactUploadException(Exception):
    """Exception raised when uploading artifacts fails."""


def upload_artifacts_via_http_api(
    *, token: str, config: ArtifactLoaderClientConfig
) -> None:
    """Upload artifacts via the HTTP API specified in the config using the provided
    token for authorization.

    Raises:
        ArtifactUploadException: If uploading artifacts fails.
    """

    event_collector = FileSystemEventCollector(config=config)
    artifacts = collect_artifacts(config=config, event_collector=event_collector)

    with httpx.Client() as client:
        response = client.post(
            f"{config.loader_api_root}/rpc/load-artifacts",
            json=artifacts,
            headers={
                "Authorization": f"Bearer {token}",
            },
            timeout=60,
        )

    if response.status_code != 204:
        raise ArtifactUploadException(
            f"Uploading artifacts failed with status code {response.status_code}."
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for storing submission metadata."""

import json
from pathlib import Path

from pydantic import BaseSettings, Field

from metldata.submission_registry import models


class SubmissionStoreConfig(BaseSettings):
    """Config parameters and their defaults."""

    submission_store_dir: Path = Field(
        ..., description="The directory where the submission JSONs will be stored."
    )


class SubmissionStore:
    """A class for storing and retrieving submissions."""

    class SubmissionDoesNotExistError(RuntimeError):
        """Raised when an Submission does not exists."""

        def __init__(self, *, submission_id: str):
            message = (
                "The submission with the following ID does not exist: " + submission_id
            )
            super().__init__(message)

    class SubmissionAlreadyExistError(RuntimeError):
        """Raised when a Submission already exists."""

        def __init__(self, *, submission_id: str):
            message = (
                "The submission with the following ID already exists: " + submission_id
            )
            super().__init__(message)

    def __init__(self, *, config: SubmissionStoreConfig):
        """Initialize with config parameters."""

        self._config = config

    def _get_submission_json_path(self, *, submission_id: str) -> Path:
        """Get the path to the JSON file containing the submission with the specified
        ID."""

        return self._config.submission_store_dir / f"{submission_id}.json"

    def _save(self, *, submission: models.Submission) -> None:
        """Save a submission to a JSON file."""

        json_path = self._get_submission_json_path(submission_id=submission.id)
        with open(json_path, "w", encoding="utf-8") as file:
            file.write(submission.json(indent=4))

    def exists(self, *, submission_id: str) -> bool:
        """Check whether a submission with the specified ID exists."""

        json_path = self._get_submission_json_path(submission_id=submission_id)
        return json_path.exists()

    def _assert_exists(self, *, submission_id: str):
        """Assert that a submission with the specified ID exists.

        Raises:
            SubmissionDoesNotExistError: Raised when the submission does not exist.
        """

        if not self.exists(submission_id=submission_id):
            raise self.SubmissionDoesNotExistError(submission_id=submission_id)

    def _assert_not_exists(self, *, submission_id: str):
        """Assert that a submission with the specified ID does not exist.

        Raises:
            SubmissionAlreadyExistError: Raised when the submission already exist.
        """

        if self.exists(submission_id=submission_id):
            raise self.SubmissionAlreadyExistError(submission_id=submission_id)

    def get_all_submission_ids(self) -> list[str]:
        """Get all submission IDs from existing submissions"""
        return sorted(
            [
                submission_path.stem
                for submission_path in self._config.submission_store_dir.iterdir()
                if submission_path.suffix == ".json"
            ]
        )

    def get_by_id(self, submission_id: str) -> models.Submission:
        """Get an existing submission by its ID.
        Raises:
            SubmissionDoesNotExistError: Raised when the submission does not exist.
        """

        json_path = self._get_submission_json_path(submission_id=submission_id)

        if not json_path.exists():
            raise self.SubmissionDoesNotExistError(submission_id=submission_id)

        with open(json_path, "r", encoding="utf-8") as file:
            submission_dict = json.load(file)

        return models.Submission(**submission_dict)

    def insert_new(self, *, submission: models.Submission) -> None:
        """Save a new submission.

        Raises:
            SubmissionAlreadyExistError: when the submission already exists.
        """

        self._assert_not_exists(submission_id=submission.id)
        self._save(submission=submission)

    def update_existing(self, *, submission: models.Submission) -> None:
        """Update an existing submission.

        Raises:
            SubmissionDoesNotExistError: when the submission does not exist.
        """

        self._assert_exists(submission_id=submission.id)
        self._save(submission=submission)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the api_factory module."""


import httpx
import pytest
from fastapi import FastAPI
from ghga_service_commons.api.testing import AsyncTestClient
from ghga_service_commons.utils.utc_dates import DateTimeUTC, now_as_utc
from hexkit.protocols.dao import DaoFactoryProtocol

from metldata.artifacts_rest.api_factory import rest_api_factory
from metldata.artifacts_rest.artifact_info import ArtifactInfo, get_artifact_info_dict
from metldata.load.aggregator import MongoDbAggregator
from metldata.load.stats import create_stats_using_aggregator
from tests.artifact_rest.test_load_artifacts import load_example_artifact_resources
from tests.fixtures.artifact_info import EXAMPLE_ARTIFACT_INFOS, MINIMAL_ARTIFACT_INFO
from tests.fixtures.mongodb import (  # noqa: F401; pylint: disable=unused-import
    MongoDbFixture,
    mongodb_fixture,
)


@pytest.mark.asyncio
async def get_example_app_client(
    dao_factory: DaoFactoryProtocol,
    artifact_infos: list[ArtifactInfo] = EXAMPLE_ARTIFACT_INFOS,
) -> httpx.AsyncClient:
    """Return a test client for a FastAPI generated using the artifact_rest_factory."""

    router = await rest_api_factory(
        artifact_infos=artifact_infos, dao_factory=dao_factory
    )

    app = FastAPI()
    app.include_router(router)
    return AsyncTestClient(app)


@pytest.mark.asyncio
async def test_artifacts_info_endpoint(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test happy path of using the artifacts info endpoint."""

    expected_infos = EXAMPLE_ARTIFACT_INFOS

    async with await get_example_app_client(
        dao_factory=mongodb_fixture.dao_factory
    ) as client:
        response = await client.options("/artifacts")

    response_json = response.json()
    assert isinstance(response_json, list)
    observed_infos = [ArtifactInfo(**info) for info in response_json]
    assert observed_infos == expected_infos


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "artifact_name, expected_info",
    [(info.name, info) for info in EXAMPLE_ARTIFACT_INFOS],
)
async def test_artifact_info_endpoint(
    artifact_name: str,
    expected_info: ArtifactInfo,
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test happy path of using the artifact info endpoint."""

    async with await get_example_app_client(
        dao_factory=mongodb_fixture.dao_factory
    ) as client:
        response = await client.options(f"/artifacts/{artifact_name}")

    observed_info = ArtifactInfo(**response.json())
    assert observed_info == expected_info


@pytest.mark.asyncio
async def test_get_artifact_resource_endpoint(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test happy path of using the get artifact resource endpoint."""

    # load example resources and prepare client:
    await load_example_artifact_resources(dao_factory=mongodb_fixture.dao_factory)

    # Get an example resource:
    artifact_name = MINIMAL_ARTIFACT_INFO.name
    class_name = "File"
    resource_id = "test_sample_01_R1"
    async with await get_example_app_client(
        dao_factory=mongodb_fixture.dao_factory,
        artifact_infos=[MINIMAL_ARTIFACT_INFO],
    ) as client:
        response = await client.get(
            f"/artifacts/{artifact_name}/classes/{class_name}/resources/{resource_id}"
        )

    assert response.status_code == 200
    observed_resource = response.json()
    # check that the right object was return using an example slot:
    expected_checksum = (
        "1c8aed294d5dec3740a175f6b655725fa668bfe41311e74f7ca9d85c91371b4e"
    )
    assert observed_resource["checksum"] == expected_checksum
    # check that the ID is included in the resource:
    assert observed_resource["alias"] == resource_id


@pytest.mark.asyncio
async def test_get_stats_endpoint(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test happy path of using the get stats endpoint."""

    # load example resources and prepare client:
    await load_example_artifact_resources(dao_factory=mongodb_fixture.dao_factory)

    artifact_infos = [MINIMAL_ARTIFACT_INFO]

    await create_stats_using_aggregator(
        artifact_infos=get_artifact_info_dict(artifact_infos=artifact_infos),
        primary_artifact_name=artifact_infos[-1].name,
        db_aggregator=MongoDbAggregator(config=mongodb_fixture.config),
    )

    # Get the global summary statistics:
    async with await get_example_app_client(
        dao_factory=mongodb_fixture.dao_factory,
        artifact_infos=artifact_infos,
    ) as client:
        response = await client.get("/stats")

    assert response.status_code == 200
    observed_stats = response.json()
    assert isinstance(observed_stats, dict)

    observed_created = DateTimeUTC.fromisoformat(observed_stats.pop("created"))
    assert abs((now_as_utc() - observed_created).seconds) < 5

    expected_stats = {
        "id": "global",
        "resource_stats": {
            "Dataset": {"count": 2},
            "File": {"count": 4, "stats": {"format": [{"value": "fastq", "count": 4}]}},
        },
    }
    assert observed_stats == expected_stats



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for resolving reference paths for existing metadata."""


from metldata.builtin_transformations.infer_references.path.path import ReferencePath
from metldata.builtin_transformations.infer_references.path.path_elements import (
    ReferencePathElement,
    ReferencePathElementType,
)
from metldata.custom_types import Json
from metldata.metadata_utils import (
    ForeignIdLookUpError,
    MetadataResourceNotFoundError,
    SelfIdLookUpError,
    lookup_foreign_ids,
    lookup_resource_by_identifier,
    lookup_self_id,
)
from metldata.model_utils.anchors import (
    AnchorPoint,
    AnchorPointNotFoundError,
    lookup_anchor_point,
)


class PathElementResolutionError(RuntimeError):
    """Raised when a path element cannot be resolved."""


def resolve_target_ids_active_element(
    *, source_resource: Json, path_element: ReferencePathElement
) -> set[str]:
    """Resolve an active reference path element applied to a metadata resource.

    Args:
        source_resource: The metadata resource to which the path element is applied.
        path_element: The active path element to resolve.

    Returns:
        A list of target IDs that are targeted by the path element in context of the
        provided source resource.

    Raises:
        PathElementResolutionError:
            if the path element cannot be resolved.
    """

    if path_element.type_ != ReferencePathElementType.ACTIVE:
        raise ValueError("Passive path element supplied where active expected.")

    try:
        target_ids = lookup_foreign_ids(
            resource=source_resource, slot=path_element.slot
        )
    except ForeignIdLookUpError as error:
        raise PathElementResolutionError(
            "Failed to resolve the path element applied to the"
            + f" resource '{source_resource}': {error}"
        ) from error

    return target_ids


def resolve_target_ids_passive_element(  # noqa: C901
    *,
    source_resource: Json,
    path_element: ReferencePathElement,
    global_metadata: Json,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> set[str]:
    """Resolve a passive reference path element applied to a metadata resource.

    Args:
        source_resource: The metadata resource to which the path element is applied.
        path_element: The passive path element to resolve.
        global_metadata: The global metadata.
        anchor_points: The anchor points by target class.

    Returns:
        A list of target IDs that are targeted by the path element in context of the
        provided source resource.

    Raises:
        PathElementResolutionError:
            if the path element cannot be resolved.
    """

    if path_element.type_ != ReferencePathElementType.PASSIVE:
        raise ValueError("Active path element supplied where passive expected.")

    try:
        source_anchor_point = lookup_anchor_point(
            class_name=path_element.source,
            anchor_points_by_target=anchor_points_by_target,
        )
    except AnchorPointNotFoundError as error:
        raise PathElementResolutionError(
            "Cannot resolve path element because of a missing anchor point for"
            + f" source class '{path_element.source}'."
        ) from error

    try:
        target_anchor_point = lookup_anchor_point(
            class_name=path_element.target,
            anchor_points_by_target=anchor_points_by_target,
        )
    except AnchorPointNotFoundError as error:
        raise PathElementResolutionError(
            "Cannot resolve path element because of a missing anchor point for"
            + f" target class '{path_element.target}'."
        ) from error

    try:
        source_identifier = lookup_self_id(
            resource=source_resource,
            identifier_slot=source_anchor_point.identifier_slot,
        )
    except SelfIdLookUpError as error:
        raise PathElementResolutionError(
            f"Cannot resolve path element: '{error}'"
        ) from error

    target_resources = global_metadata.get(target_anchor_point.root_slot)
    if target_resources is None:
        raise PathElementResolutionError(
            "Cannot resolve path element: No target resources found for"
            + f" root slot '{target_anchor_point.root_slot}'."
        )

    # lookup the target resources:
    target_ids_of_interest: set[str] = set()
    for target_resource in target_resources:
        referenced_source_ids = lookup_foreign_ids(
            resource=target_resource, slot=path_element.slot
        )

        if source_identifier in referenced_source_ids:
            target_id = lookup_self_id(
                resource=target_resource,
                identifier_slot=target_anchor_point.identifier_slot,
            )
            target_ids_of_interest.add(target_id)

    return target_ids_of_interest


def resolve_path_element(
    *,
    source_resource: Json,
    global_metadata: Json,
    path_element: ReferencePathElement,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> list[Json]:
    """Resolve a reference path element applied to a metadata resource.

    Returns:
        A list of metadata resources that are targeted by the path element.

    Raises:
        PathElementResolutionError:
            if the path element cannot be resolved.
    """

    if path_element.type_ == ReferencePathElementType.ACTIVE:
        target_ids = resolve_target_ids_active_element(
            source_resource=source_resource, path_element=path_element
        )
    else:
        target_ids = resolve_target_ids_passive_element(
            source_resource=source_resource,
            path_element=path_element,
            global_metadata=global_metadata,
            anchor_points_by_target=anchor_points_by_target,
        )

    if not target_ids:
        return []

    target_resources: list[Json] = []
    for target_id in target_ids:
        try:
            target_resource = lookup_resource_by_identifier(
                class_name=path_element.target,
                identifier=target_id,
                global_metadata=global_metadata,
                anchor_points_by_target=anchor_points_by_target,
            )
        except MetadataResourceNotFoundError as error:
            raise PathElementResolutionError(
                f"Cannot resolve path element for source resource '{source_resource}'"
                + f" because the target resource with ID '{target_id}' could not be"
                + " found."
            ) from error
        target_resources.append(target_resource)

    return target_resources


def resolve_reference_for_metadata_resource(
    *,
    resource: Json,
    global_metadata: Json,
    reference_path: ReferencePath,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> list[Json]:
    """Resolve an inferred reference for an individual metadata resource.

    Args:
        resource: The metadata resource to resolve the reference for.
        global_metadata: The global metadata context to look up references in.
        reference_path: The path of the inferred reference.
        anchor_points: The anchor points of the metadata model.

    Returns:
        A list of metadata resources that are targeted by the reference.

    Raises:
        PathElementResolutionError:
            if the reference resolution fails.
    """

    source_resources = [resource]
    for path_element in reference_path.elements:
        target_resources: list[Json] = []
        for source_resource in source_resources:
            local_target_resources = resolve_path_element(
                source_resource=source_resource,
                global_metadata=global_metadata,
                path_element=path_element,
                anchor_points_by_target=anchor_points_by_target,
            )
            target_resources.extend(local_target_resources)

        # set the target resources of the current path element as the source resources
        # for the next path element:
        source_resources = target_resources

    return source_resources



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the load_artifacts module"""

import pytest
from hexkit.protocols.dao import DaoFactoryProtocol

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.load_resources import load_artifact_resources
from metldata.artifacts_rest.models import ArtifactResource
from tests.fixtures.artifact_info import MINIMAL_ARTIFACT_INFO
from tests.fixtures.metadata import VALID_MINIMAL_METADATA_EXAMPLE
from tests.fixtures.mongodb import (  # noqa: F401; pylint: disable=unused-import
    MongoDbFixture,
    mongodb_fixture,
)


async def load_example_artifact_resources(
    dao_factory: DaoFactoryProtocol,
) -> ArtifactDaoCollection:
    """Load the example artifact using the load_artifact_resources function and
    returns a ArtifactDaoCollection for accessing the resources."""

    # construct the dao collection:
    dao_collection = await ArtifactDaoCollection.construct(
        dao_factory=dao_factory,
        artifact_infos=[MINIMAL_ARTIFACT_INFO],
    )

    # load artifact resources from the metadata example:
    await load_artifact_resources(
        artifact_content=VALID_MINIMAL_METADATA_EXAMPLE,
        artifact_info=MINIMAL_ARTIFACT_INFO,
        dao_collection=dao_collection,
    )

    return dao_collection


@pytest.mark.asyncio
async def test_load_artifact_resources(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test happy path of using load_artifact_resources function."""

    dao_collection = await load_example_artifact_resources(
        dao_factory=mongodb_fixture.dao_factory,
    )

    # check that artifact resources have been persisted to the database by testing for
    # an example file:
    expected_resource = ArtifactResource(
        id_="test_sample_01_R1",
        class_name="File",
        content={
            "alias": "test_sample_01_R1",
            "filename": "test_sample_01_R1.fastq",
            "format": "fastq",
            "checksum": "1c8aed294d5dec3740a175f6b655725fa668bfe41311e74f7ca9d85c91371b4e",
            "size": 299943,
        },
    )
    dao = await dao_collection.get_dao(
        artifact_name=MINIMAL_ARTIFACT_INFO.name,
        class_name=expected_resource.class_name,
    )
    observed_resource = await dao.get_by_id(id_=expected_resource.id_)
    assert observed_resource == expected_resource



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for publishing artifacts."""


import json

from hexkit.protocols.eventpub import EventPublisherProtocol
from pydantic import BaseModel

from metldata.event_handling.artifact_events import (
    ArtifactEventConfig,
    get_artifact_topic,
)
from metldata.event_handling.models import SubmissionEventPayload


class ArtifactEventPublisherConfig(ArtifactEventConfig):
    """Config parameters and their defaults."""


class ArtifactEvent(BaseModel):
    """Artifact event model."""

    artifact_type: str
    payload: SubmissionEventPayload


class ArtifactEventPublisher:
    """Handles publication of artifact events."""

    def __init__(
        self, *, config: ArtifactEventPublisherConfig, provider: EventPublisherProtocol
    ):
        """Initialize with config parameters."""

        self._config = config
        self._provider = provider

    async def publish_artifact(self, artifact_event: ArtifactEvent):
        """Publish an artifact as submission event"""

        payload = json.loads(artifact_event.payload.json())
        topic = get_artifact_topic(
            artifact_topic_prefix=self._config.artifact_topic_prefix,
            artifact_type=artifact_event.artifact_type,
        )
        type_ = artifact_event.artifact_type

        await self._provider.publish(
            topic=topic,
            type_=type_,
            key=artifact_event.payload.submission_id,
            payload=payload,
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling reference paths."""

from metldata.builtin_transformations.infer_references.path.path_str import (
    PATH_PATTERN,
    ValidationError,
    clean_path_str,
    path_str_to_object_elements,
)


class ReferencePath:
    """A model describing the path of a reference between classes of a metadata model.

    The reference path has two available representation. A string-based ("path_str"
    attribute) and an element-based ("elements" attribute) one.

    In the string-based representation ("path_string" attribute), the first and the last
    word correspond the name of the source and target class, respectively. ">" and "<"
    means indicate active (left class references the right one) and passive (the left
    class is referenced by the right one). Parentheses attached to these angles thereby
    indicate the slot name of the referencing class. E.g. "class_a(class_b)>class_b"
    means that the source class "class_a" has a slot "class_b" that references the
    target class "class_b". Or "class_a<(class_a)class_b" means that the source
    class "class_a" is reference by the target class "class_b" via its slots
    "class_a". Reference paths can also involve additional classes. E.g. a string of
    "class_a<(class_a)class_b(class_c)>class_c" means that
    a reference from the source class "class_a" to the target class "class_c" can be
    established via an additional class "class_b". Any inserted spaces or newlines will
    be ignored. So the following paths are equivalent:
        - "class_a (class_b)> class_b"
        - "class_a
            (class_b)>
            class_b"

    A reference path consists of one or more elements. An element is a relationship
    between two classes. Reference paths that establish a direct relationship between
    source and target classes without the use of additional classes have only one
    element (e.g. in string reprentations "class_a(class_b)>class_b" or
    "class_a<(class_a)class_b"). More complex paths consist of multiple elements.
    E.g. the path "class_a<(class_a)class_b(class_c)>class_c" can be decomposed
    into the elements: "class_a<(class_a)class_b" and
    "class_b(class_c)>class_c".

    The elements of a ReferencePath are stored in the "elements" attribute as a list
    of ReferencePathElement objects that are optimized for programmatic use.

    The "source" attribute provides the source class of the path while the
    "target" attribute provides the target class of the path.
    """

    def __init__(self, *, path_str: str):
        """Construct reference path from a string-based representation."""

        self.path_str = clean_path_str(path_str=path_str)
        self.elements = path_str_to_object_elements(path_str=self.path_str)
        self.source = self.elements[0].source
        self.target = self.elements[-1].target

    @classmethod
    def validate(cls, value) -> "ReferencePath":
        """A validator for pydantic."""

        if isinstance(value, cls):
            return value

        if not isinstance(value, str):
            raise ValueError("A string is required.")

        try:
            return cls(path_str=value)
        except ValidationError as error:
            raise ValueError(str(error)) from ValidationError

    @classmethod
    def __get_validators__(cls):
        """To get validators for pydantic"""

        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema: dict):
        """Modify the field schema for pydantic."""

        field_schema.update(type="string", pattern=PATH_PATTERN)

    def __eq__(self, other: object):
        """For comparisons."""

        if not isinstance(other, ReferencePath):
            return NotImplemented

        return self.path_str == other.path_str



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A transformation that normalizes a model to a canonical form."""

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model.meta import SlotDefinition

from metldata.model_utils.essentials import MetadataModel


def resolve_inheritance(output_model: dict, input_schema_view: SchemaView) -> None:
    """Resolve inheritance by embedding all slots into the derived classes."""
    all_slot_names = set()
    for cls_name, cls in output_model["classes"].items():
        slot_usages = input_schema_view.class_induced_slots(cls_name)
        slot_names = [slot.name for slot in slot_usages]
        # Integrate all slot information
        cls["slot_usage"] = {slot_def.name: slot_def for slot_def in slot_usages}
        cls["slots"] = slot_names
        # Update all_slot names for globally reduced set of actually used slot
        # names.
        all_slot_names.update(slot_names)
        # Remove inheritance
        cls["is_a"] = None
        cls["mixins"] = []
        cls["attributes"] = []
    # Rewrite slots
    output_model["slots"] = {
        slot_name: SlotDefinition(name=slot_name) for slot_name in all_slot_names
    }


def embed_types(output_model: dict, input_schema_view: SchemaView) -> None:
    """Embed types from the types schema into the model."""
    if "linkml:types" in output_model["imports"]:
        # Load the types schema
        types_schema = input_schema_view.load_import("linkml:types")
        # Embed types
        if not isinstance(types_schema.types, dict):
            raise RuntimeError("Types schema does not contain a dict of types.")
        for type_name, type_def in types_schema.types.items():
            if type_name not in output_model["types"]:
                output_model["types"][type_name] = type_def
        # Embed prefixes
        if not isinstance(types_schema.prefixes, dict):
            raise RuntimeError("Types schema does not contain a dict of types.")
        for prefix_name, prefix_def in types_schema.prefixes.items():
            if prefix_name not in output_model["prefixes"]:
                output_model["prefixes"][prefix_name] = prefix_def
        # Drop import
        output_model["imports"].remove("linkml:types")


def normalize_model(model: MetadataModel) -> MetadataModel:
    """Normalize model to canonical form with all slots being globally defined
    only as empty stubs and all slot definitions being defined in the slot_usage
    of the respective class."""

    output_model = model.as_dict()
    schema_view = model.schema_view

    resolve_inheritance(output_model, schema_view)
    embed_types(output_model, schema_view)

    return MetadataModel(**output_model)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling artifact events."""

from pydantic import BaseModel, Field, validator


class ArtifactEventConfig(BaseModel):
    """Config parameters and their defaults."""

    artifact_topic_prefix: str = Field(
        ...,
        description=(
            "The prefix used for topics containing artifacts. The topic name is"
            + " expected to be '{prefix}.{artifact_type}'. The prefix must not contain"
            + " dots."
        ),
    )

    # pylint: disable=no-self-argument
    @validator("artifact_topic_prefix")
    def artifact_topic_prefix_must_not_contain_dots(cls, value: str):
        """Validate that artifact topic prefix does not contain dots."""

        if "." in value:
            raise ValueError(f"Artifact topic prefix '{value}' must not contain dots.")

        return value


def get_artifact_topic(*, artifact_topic_prefix: str, artifact_type: str) -> str:
    """Get the topic name for the given artifact type.
    Both the provided artifact_topic_prefix and the artifact_type must not contain dots.
    """

    if "." in artifact_topic_prefix:
        raise ValueError(
            f"Artifact topic prefix '{artifact_topic_prefix}' must not contain dots."
        )
    if "." in artifact_type:
        raise ValueError(f"Artifact type '{artifact_type}' must not contain dots.")

    return f"{artifact_topic_prefix}.{artifact_type}"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic to build REST APIs for querying artifacts."""

from fastapi import HTTPException
from fastapi.routing import APIRouter
from hexkit.protocols.dao import DaoFactoryProtocol, ResourceNotFoundError

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.artifact_info import get_artifact_info_dict
from metldata.artifacts_rest.models import ArtifactInfo, GlobalStats
from metldata.artifacts_rest.query_resources import (
    ArtifactResourceNotFoundError,
    query_artifact_resource,
)
from metldata.custom_types import Json
from metldata.load.stats import STATS_COLLECTION_NAME


async def rest_api_factory(
    *, artifact_infos: list[ArtifactInfo], dao_factory: DaoFactoryProtocol
) -> APIRouter:
    """Factory to build a REST API for querying the provided artifact.

    Args:
        artifact_info: Information on the artifact to build the REST API for.
        dao_factory: An implementation of the DaoFactoryProtocol.

    Returns:
        A FastAPI router with RESTful endpoints to query artifacts.
    """

    artifact_info_dict = get_artifact_info_dict(artifact_infos=artifact_infos)
    dao_collection = await ArtifactDaoCollection.construct(
        dao_factory=dao_factory, artifact_infos=artifact_infos
    )
    stats_dao = await dao_factory.get_dao(
        name=STATS_COLLECTION_NAME, dto_model=GlobalStats, id_field="id"
    )

    router = APIRouter()

    @router.options("/artifacts")
    async def get_artifacts_info() -> list[ArtifactInfo]:
        """Get information on available artifacts."""

        return artifact_infos

    @router.options("/artifacts/{artifact_name}")
    async def get_artifact_info(artifact_name: str) -> ArtifactInfo:
        """Get information on the artifact with the given name."""

        artifact_info = artifact_info_dict.get(artifact_name)
        if artifact_info is None:
            raise HTTPException(status_code=404, detail="Artifact not found.")

        return artifact_info

    @router.get(
        "/artifacts/{artifact_name}/classes/{class_name}/resources/{resource_id}"
    )
    async def get_artifact_resource(
        artifact_name: str, class_name: str, resource_id: str
    ) -> Json:
        """Get the resource with the given id of the class with the given name from the
        artifact with the given name."""

        try:
            resource = await query_artifact_resource(
                artifact_name=artifact_name,
                class_name=class_name,
                resource_id=resource_id,
                dao_collection=dao_collection,
            )
        except ArtifactResourceNotFoundError as error:
            raise HTTPException(status_code=404, detail=str(error)) from error

        return resource.content

    @router.get("/stats")
    async def get_stats() -> GlobalStats:
        """Get the global summary statistics for all resources."""

        try:
            resource = await stats_dao.get_by_id("global")
        except ResourceNotFoundError as error:
            raise HTTPException(status_code=404, detail=str(error)) from error

        return resource

    return router



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Metadata models."""

from metldata.model_utils.essentials import MetadataModel
from tests.fixtures.utils import BASE_DIR

EXAMPLE_MODEL_DIR = BASE_DIR / "example_models"
VALID_MINIMAL_MODEL_EXAMPLE_PATH = EXAMPLE_MODEL_DIR / "minimal_model.yaml"


def _get_example_model(name: str) -> MetadataModel:
    """Get example model."""

    return MetadataModel.init_from_path(EXAMPLE_MODEL_DIR / f"{name}.yaml")


VALID_MINIMAL_METADATA_MODEL = _get_example_model("minimal_model")
VALID_ADVANCED_METADATA_MODEL = _get_example_model("advanced_model")
VALID_METADATA_MODELS = [VALID_MINIMAL_METADATA_MODEL, VALID_ADVANCED_METADATA_MODEL]

ROOT_CLASS_INVALID_MODELS = [
    _get_example_model(f"minimal_model_root_class_missing{idx}") for idx in range(1, 3)
]

ANCHORS_INVALID_MODELS = [
    _get_example_model(f"minimal_model_anchors_invalid{idx}") for idx in range(1, 5)
]

INVALID_METADATA_MODELS = ROOT_CLASS_INVALID_MODELS + ANCHORS_INVALID_MODELS



#!/usr/bin/env python3

# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Run all update scripts that are present in the repository in the correct order"""

try:
    from scripts.update_template_files import main as update_template
except ImportError:
    pass
else:
    print("Pulling in updates from template repository")
    update_template()

try:
    from scripts.update_config_docs import main as update_config
except ImportError:
    pass
else:
    print("Updating config docs")
    update_config()

try:
    from scripts.update_openapi_docs import main as update_openapi
except ImportError:
    pass
else:
    print("Updating OpenAPI docs")
    update_openapi()

try:
    from scripts.update_readme import main as update_readme
except ImportError:
    pass
else:
    print("Updating README")
    update_readme()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Config Parameter Modeling and Parsing"""

from hexkit.config import config_from_yaml

# pylint: disable=unused-import
from metldata.accession_registry.config import Config as AccessionRegistryConfig
from metldata.artifacts_rest.config import ArtifactsRestConfig

# pylint: disable=unused-import
from metldata.load.config import ArtifactLoaderAPIConfig
from metldata.submission_registry.config import Config as SubmissionRegistryConfig


# pylint: disable=too-many-ancestors
@config_from_yaml(prefix="metl_sub")
class SubmissionConfig(AccessionRegistryConfig, SubmissionRegistryConfig):
    """Config parameters and their defaults."""


@config_from_yaml(prefix="metldata")
class Config(ArtifactLoaderAPIConfig, ArtifactsRestConfig):
    """Config parameters and their defaults."""

    service_name: str = "metldata"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""API for loading artifacts into running services."""

from typing import Optional

from fastapi import Depends, HTTPException, Response, Security
from fastapi.routing import APIRouter
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from ghga_service_commons.auth.context import AuthContextProtocol
from ghga_service_commons.auth.policies import require_auth_context_using_credentials
from hexkit.protocols.dao import DaoFactoryProtocol
from hexkit.providers.akafka import KafkaEventPublisher
from pydantic import BaseModel, Field

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.artifact_info import get_artifact_info_dict
from metldata.artifacts_rest.models import ArtifactInfo
from metldata.load.aggregator import DbAggregator
from metldata.load.auth import check_token
from metldata.load.config import ArtifactLoaderAPIConfig
from metldata.load.event_publisher import EventPubTranslator
from metldata.load.load import (
    ArtifactResourcesInvalid,
    check_artifact_resources,
    load_artifacts_using_dao,
)
from metldata.load.models import ArtifactResourceDict
from metldata.load.stats import create_stats_using_aggregator


class LoaderTokenAuthContext(BaseModel):
    """An auth context used to contain the loader token."""

    token: str = Field(
        ...,
        description="A simple alphanumeric token to authenticate for loading artifacts.",
    )


class LoaderTokenAuthProvider(AuthContextProtocol[LoaderTokenAuthContext]):
    """A provider for the loader token auth context."""

    def __init__(self, *, token_hashes: list[str]):
        """Initialize the loader token auth provider."""

        self._token_hashes = token_hashes

    async def get_context(self, token: str) -> Optional[LoaderTokenAuthContext]:
        """Get a loader token auth context."""

        if not token:
            return None

        if not check_token(token=token, token_hashes=self._token_hashes):
            raise self.AuthContextValidationError("Invalid token.")

        return LoaderTokenAuthContext(token=token)


async def rest_api_factory(
    *,
    artifact_infos: list[ArtifactInfo],
    primary_artifact_name: str,
    config: ArtifactLoaderAPIConfig,
    dao_factory: DaoFactoryProtocol,
    db_aggregator: DbAggregator,
    token_hashes: list[str],
) -> APIRouter:
    """Return a router for an API for loading artifacts."""

    artifact_info_dict = get_artifact_info_dict(artifact_infos=artifact_infos)
    dao_collection = await ArtifactDaoCollection.construct(
        dao_factory=dao_factory, artifact_infos=artifact_infos
    )

    async def require_token_context(
        credentials: HTTPAuthorizationCredentials = Depends(
            HTTPBearer(auto_error=True)
        ),
    ) -> LoaderTokenAuthContext:
        """Require a VIP authentication and authorization context using FastAPI."""
        return await require_auth_context_using_credentials(
            credentials=credentials,
            auth_provider=LoaderTokenAuthProvider(token_hashes=token_hashes),
        )

    require_token = Security(require_token_context)

    router = APIRouter()

    @router.post("/rpc/load-artifacts")
    async def load_artifacts(
        artifact_resources: ArtifactResourceDict,
        _token: LoaderTokenAuthContext = require_token,
    ):
        """Load artifacts into services for querying via user-accessible API."""

        try:
            check_artifact_resources(
                artifact_resources=artifact_resources,
                artifact_infos=artifact_info_dict,
            )
        except ArtifactResourcesInvalid as error:
            raise HTTPException(status_code=422, detail=str(error)) from error

        async with KafkaEventPublisher.construct(config=config) as event_pub_provider:
            event_publisher = EventPubTranslator(
                config=config, provider=event_pub_provider
            )
            await load_artifacts_using_dao(
                artifact_resources=artifact_resources,
                artifact_info_dict=artifact_info_dict,
                event_publisher=event_publisher,
                dao_collection=dao_collection,
            )

            await create_stats_using_aggregator(
                artifact_infos=artifact_info_dict,
                primary_artifact_name=primary_artifact_name,
                db_aggregator=db_aggregator,
            )

            return Response(status_code=204)

    return router



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling information on artifacts."""

import json
from typing import Optional

from linkml.generators.jsonschemagen import JsonSchemaGenerator
from linkml_runtime.linkml_model.annotations import Annotation
from linkml_runtime.linkml_model.meta import ClassDefinition

from metldata.artifacts_rest.models import ArtifactInfo, ArtifactResourceClass
from metldata.custom_types import Json
from metldata.model_utils.anchors import (
    get_anchors_points_by_target,
    lookup_anchor_point,
)
from metldata.model_utils.assumptions import check_basic_model_assumption
from metldata.model_utils.essentials import MetadataModel


def is_class_hidden(*, class_definition: ClassDefinition) -> bool:
    """Check if a metadata class is annotated as `hidden`."""

    if not isinstance(class_definition.annotations, dict):
        raise RuntimeError(  # This should never happen
            f"Annotations of class {class_definition.name} are not a dictionary."
        )

    if "hidden" in class_definition.annotations:
        if not isinstance(class_definition.annotations["hidden"], Annotation):
            raise RuntimeError(  # This should never happen
                f"Annotation 'hidden' of class {class_definition.name} has an"
                + " unexpected format."
            )

        if not isinstance(class_definition.annotations["hidden"].value, bool):
            raise RuntimeError(  # This should never happen
                f"Annotation 'hidden' of class {class_definition.name} has an"
                + " unexpected value."
            )

        return class_definition.annotations["hidden"].value

    return False


def get_resource_class_names(*, model: MetadataModel) -> list[str]:
    """Get the names of all classes from a metadata model.
    Only anchored classes are considered. Classes that are annotated as `hidden`
    are ignored.
    """

    anchored_class_names = get_anchors_points_by_target(model=model).keys()
    schema_view = model.schema_view

    return [
        anchored_class_name
        for anchored_class_name in anchored_class_names
        if not is_class_hidden(
            class_definition=schema_view.get_class(anchored_class_name, strict=True)
        )
    ]


def filter_json_schema_definitions(
    *, target_definition: Json, all_definitions: dict[str, Json]
) -> dict[str, Json]:
    """Remove JSON schema definitions that are not used by the target_defintion."""

    target_definition_str = json.dumps(target_definition)
    return {
        definition_name: definition_schema
        for definition_name, definition_schema in all_definitions.items()
        if f'"$ref": "#/$defs/{definition_name}"' in target_definition_str
    }


def subset_json_schema_for_class(*, global_json_schema: Json, class_name: str) -> Json:
    """Subset a global json schema for the specified class.
    Please note, the resulting schema might contain definitions that are not used.
    """

    definitions = global_json_schema.get("$defs")
    if not definitions:
        raise RuntimeError(  # This should never happen
            "Schema does not contain definitions."
        )

    target_definition = definitions.get(class_name)
    if not target_definition:
        raise RuntimeError(  # This should never happen
            f"Schema does not contain definition for class {class_name}."
        )

    filtered_definitions = filter_json_schema_definitions(
        target_definition=target_definition, all_definitions=definitions
    )

    return {
        "$schema": global_json_schema["$schema"],
        **target_definition,
        "$defs": filtered_definitions,
    }


def load_description_for_class(
    *, model: MetadataModel, class_name: str
) -> Optional[str]:
    """Load the description for the specified class of the metadata model."""

    return model.schema_view.get_class(class_name, strict=True).description


def load_resource_classes(*, model: MetadataModel) -> dict[str, ArtifactResourceClass]:
    """Load all classes from a metadata model.
    Only anchored classes are considered. Classes that are annotated as `hidden`
    are ignored.

    Args:
        model: The metadata model.

    Returns:
        A dictionary of resource classes for this artifact.
        The keys are the names of the metadata classes. The values are the
        corresponding metadata class models.
    """

    anchor_points_by_target = get_anchors_points_by_target(model=model)

    resource_class_names = get_resource_class_names(model=model)

    global_json_schema = json.loads(JsonSchemaGenerator(model).serialize())

    return {
        resource_class_name: ArtifactResourceClass(
            name=resource_class_name,
            description=load_description_for_class(
                model=model, class_name=resource_class_name
            ),
            anchor_point=lookup_anchor_point(
                class_name=resource_class_name,
                anchor_points_by_target=anchor_points_by_target,
            ),
            json_schema=subset_json_schema_for_class(
                global_json_schema=global_json_schema, class_name=resource_class_name
            ),
        )
        for resource_class_name in resource_class_names
    }


def load_artifact_info(
    *, name: str, description: str, model: MetadataModel
) -> ArtifactInfo:
    """Load artifact info from a metadata model.

    Args:
        name: The name of the artifact.
        description: A description of the artifact.
        model: The metadata model for the artifact.

    Returns:
        The artifact info model.
    """

    check_basic_model_assumption(model)

    resource_classes = load_resource_classes(model=model)

    return ArtifactInfo(
        name=name,
        description=description,
        resource_classes=resource_classes,
    )


def get_artifact_info_dict(
    *, artifact_infos: list[ArtifactInfo]
) -> dict[str, ArtifactInfo]:
    """Build a dictionary from artifact name to artifact info."""

    # check if artifact names are unique:
    artifact_names = [artifact_info.name for artifact_info in artifact_infos]
    if len(artifact_names) != len(set(artifact_names)):
        raise ValueError("Artifact names must be unique.")

    return {artifact_info.name: artifact_info for artifact_info in artifact_infos}



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the main modules."""

import json
from copy import deepcopy

import pytest
from ghga_service_commons.utils.utc_dates import now_as_utc
from hexkit.protocols.dao import ResourceNotFoundError
from hexkit.providers.akafka.testutils import (  # noqa: F401; pylint: disable=unused-import
    KafkaFixture,
    kafka_fixture,
)

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.load_resources import (
    convert_file_information,
    get_file_slots,
)
from metldata.artifacts_rest.models import ArtifactResource, GlobalStats
from metldata.load.auth import generate_token
from metldata.load.stats import STATS_COLLECTION_NAME
from tests.fixtures.load.joint import (  # noqa: F401; pylint: disable=unused-import
    EMBEDDED_DATASET_TEST_PATH,
    JointFixture,
    joint_fixture,
)
from tests.fixtures.mongodb import (  # noqa: F401; pylint: disable=unused-import
    MongoDbFixture,
    mongodb_fixture,
)
from tests.fixtures.workflows import EXAMPLE_ARTIFACTS


@pytest.mark.asyncio
async def test_load_artifacts_endpoint_happy(
    joint_fixture: JointFixture,  # noqa: F811
):
    """Test the happy path of using the load artifacts endpoint."""

    async with joint_fixture.kafka.record_events(
        in_topic=joint_fixture.config.dataset_change_event_topic
    ) as dataset_recorder:
        async with joint_fixture.kafka.record_events(
            in_topic=joint_fixture.config.resource_change_event_topic
        ) as resource_recorder:
            response = await joint_fixture.client.post(
                "/rpc/load-artifacts",
                json=joint_fixture.artifact_resources,
                headers={"Authorization": f"Bearer {joint_fixture.token}"},
            )
            assert response.status_code == 204

    assert len(resource_recorder.recorded_events) == 2
    for event in resource_recorder.recorded_events:
        assert event.type_ == joint_fixture.config.resource_upsertion_type
    assert len(dataset_recorder.recorded_events) == 2
    for event in dataset_recorder.recorded_events:
        assert event.type_ == joint_fixture.config.dataset_upsertion_type

    # check that the artifact resources were loaded based on an example:
    expected_artifact_name = "embedded_public"

    expected_file_resource_class = "StudyFile"
    expected_file_resource_id = joint_fixture.expected_file_resource_content[
        "accession"
    ]

    expected_embedded_dataset_resource_class = "EmbeddedDataset"
    expected_embedded_dataset_resource_id = (
        joint_fixture.expected_embedded_dataset_resource_content["accession"]
    )

    dao_collection = await ArtifactDaoCollection.construct(
        dao_factory=joint_fixture.mongodb.dao_factory,
        artifact_infos=joint_fixture.artifact_infos,
    )

    dao = await dao_collection.get_dao(
        artifact_name=expected_artifact_name, class_name=expected_file_resource_class
    )
    observed_resource = await dao.get_by_id(expected_file_resource_id)
    assert observed_resource.content == joint_fixture.expected_file_resource_content

    dao = await dao_collection.get_dao(
        artifact_name=expected_artifact_name,
        class_name=expected_embedded_dataset_resource_class,
    )
    observed_resource = await dao.get_by_id(expected_embedded_dataset_resource_id)
    assert (
        observed_resource.content
        == joint_fixture.expected_embedded_dataset_resource_content
    )

    # check that the summary statistics has been created:
    expected_resource_stats = {
        "DataAccessPolicy": {"count": 1},
        "Dataset": {"count": 2},
        "DataAccessCommittee": {"count": 1},
        "StudyFile": {
            "count": 3,
            "stats": {"format": [{"value": "FASTQ", "count": 3}]},
        },
        "Study": {"count": 1},
        "EmbeddedDataset": {"count": 2},
    }
    stats_dao = await joint_fixture.mongodb.dao_factory.get_dao(
        name=STATS_COLLECTION_NAME, dto_model=GlobalStats, id_field="id"
    )
    async for stats in stats_dao.find_all(mapping={}):
        assert stats.id == "global"
        assert abs((now_as_utc() - stats.created).seconds) < 5
        assert stats.resource_stats == expected_resource_stats

    # replace tested resource with slightly changed one
    changed_accession = "CHANGED_EMBEDDED_DATASET"
    new_artifact_resources = deepcopy(joint_fixture.artifact_resources)
    del new_artifact_resources[expected_artifact_name][0]["embedded_dataset"][0]
    new_artifact_resources[expected_artifact_name][0]["embedded_dataset"][0][
        "accession"
    ] = "CHANGED_EMBEDDED_DATASET"
    expected_resource_content = new_artifact_resources[expected_artifact_name][0][
        "embedded_dataset"
    ][0]

    # submit changed request
    async with joint_fixture.kafka.record_events(
        in_topic=joint_fixture.config.dataset_change_event_topic
    ) as dataset_recorder:
        async with joint_fixture.kafka.record_events(
            in_topic=joint_fixture.config.resource_change_event_topic
        ) as resource_recorder:
            response = await joint_fixture.client.post(
                "/rpc/load-artifacts",
                json=new_artifact_resources,
                headers={"Authorization": f"Bearer {joint_fixture.token}"},
            )
            assert response.status_code == 204

    assert len(resource_recorder.recorded_events) == 3
    for event in resource_recorder.recorded_events:
        if event.key == f"dataset_embedded_{changed_accession}":
            assert event.type_ == joint_fixture.config.resource_upsertion_type
        else:
            assert event.type_ == joint_fixture.config.resource_deletion_event_type

    assert len(dataset_recorder.recorded_events) == 3

    for event in dataset_recorder.recorded_events:
        if event.key == f"dataset_embedded_{changed_accession}":
            assert event.type_ == joint_fixture.config.dataset_upsertion_type
        else:
            assert event.type_ == joint_fixture.config.dataset_deletion_type

    observed_resource = await dao.get_by_id(changed_accession)
    assert observed_resource.content == expected_resource_content

    # submit an empty request:
    async with joint_fixture.kafka.record_events(
        in_topic=joint_fixture.config.dataset_change_event_topic
    ) as dataset_recorder:
        async with joint_fixture.kafka.record_events(
            in_topic=joint_fixture.config.resource_change_event_topic
        ) as resource_recorder:
            response = await joint_fixture.client.post(
                "/rpc/load-artifacts",
                json={},
                headers={"Authorization": f"Bearer {joint_fixture.token}"},
            )
            assert response.status_code == 204

    assert len(resource_recorder.recorded_events) == 9
    for event in resource_recorder.recorded_events:
        assert event.type_ == joint_fixture.config.resource_deletion_event_type

    assert len(dataset_recorder.recorded_events) == 1
    for event in dataset_recorder.recorded_events:
        assert event.type_ == joint_fixture.config.dataset_deletion_type

    # confirm that example resource was deleted:
    with pytest.raises(ResourceNotFoundError):
        await dao.get_by_id(changed_accession)


@pytest.mark.asyncio
async def test_load_artifacts_endpoint_invalid_resources(
    joint_fixture: JointFixture,  # noqa: F811
):
    """Test using the load artifacts endpoint with resources of unknown artifacts."""

    # load example artifacts resources:
    unknown_artifact_resources = {
        "unknown_artifact": [list(EXAMPLE_ARTIFACTS.values())[0]]
    }
    response = await joint_fixture.client.post(
        "/rpc/load-artifacts",
        json=unknown_artifact_resources,
        headers={"Authorization": f"Bearer {joint_fixture.token}"},
    )
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_load_artifacts_endpoint_invalid_token(
    joint_fixture: JointFixture,  # noqa: F811
):
    """Test that using the load artifacts endpoint with an invalid token fails."""

    invalid_token = generate_token()

    # load artifact resources with invalid token:
    response = await joint_fixture.client.post(
        "/rpc/load-artifacts",
        json={},
        headers={"Authorization": f"Bearer {invalid_token}"},
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_file_slot_transformation(
    joint_fixture: JointFixture,  # noqa: F811
):
    """Test that file slots are discovered and file extensions are extracted correctly"""

    with open(EMBEDDED_DATASET_TEST_PATH, "r", encoding="utf-8") as file:
        embedded_datasets = json.load(file)

    file_slots = []
    artifact_info = joint_fixture.artifact_infos[0]

    # get all embedded datset resources
    for embedded_dataset in embedded_datasets["embedded_dataset"]:
        resource = ArtifactResource(
            id_=embedded_dataset["accession"],
            class_name="EmbeddedDataset",
            content=embedded_dataset,
        )
        current_file_slots = get_file_slots(
            artifact_info=artifact_info, resource=resource
        )
        file_slots.extend(current_file_slots)

    extension_mapping = {
        "STUDY_FILE_1": ".fastq.gz",
        "STUDY_FILE_2": ".fastq",
        "STUDY_FILE_3": ".gz",
        "STUDY_FILE_4": "",
    }
    for file_info in convert_file_information(file_slots=file_slots):
        assert extension_mapping[file_info.accession] == file_info.file_extension



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for defining embedding profiles."""

from __future__ import annotations

from typing import Optional, Union

from pydantic import BaseModel, Field


class EmbeddingProfile(BaseModel):
    """A model for describing a profile for embedding referenced classes into a class
    of interest of a metadata model. Please note, only the embedding for anchored
    classes that are referenced by this source class can be changed. All anchored
    classes are assumed to be non-embedded by default. The embedding profile can be used
    to define anchored classes as embedded given the slot named used for renferencing
    in the source class.
    """

    target_class: str = Field(
        ..., description="The name of the transformed class with embeddings."
    )
    source_class: str = Field(
        ..., description="The class to which the this embedding profile applies."
    )
    description: Optional[str] = Field(
        ..., description="Description of the transformed class with embeddings."
    )
    embedded_references: dict[str, Union[str, EmbeddingProfile]] = Field(
        ...,
        description=(
            "The references embedded into the target class."
            + "The keys are the names of slots in the target class that are used for "
            + " the references to other classes. The values are either the names of the"
            + " referenced classes or other embedding profiles if a custom embedding"
            + " will be applied to the referenced classes, too."
        ),
    )
